__version__ = '1.1.4'
__license__ = 'MIT License'
__docsurl__ = 'https://imgmarker.readthedocs.io/en/latest/'

import sys
import os
from importlib import resources

def resource_path(resource):
    if hasattr(sys,'_MEIPASS'):
        return os.path.join(sys._MEIPASS, resource)
    else: 
        return str(resources.files(__package__).joinpath(resource))

ICON = resource_path('icon.ico')
HEART_SOLID = resource_path('heart_solid.ico')
HEART_CLEAR = resource_path('heart_clear.ico')



"""
Copyright © 2025, UChicago Argonne, LLC

Full license found at _YOUR_INSTALLATION_DIRECTORY_/imgmarker/LICENSE
"""

"""This module contains the `Mark` class and related classes."""

from imgmarker.gui.pyqt import QGraphicsPathItem, QPainterPath, QGraphicsProxyWidget, QLineEdit, QPen, QColor, Qt, QPointF, QEvent
from imgmarker import config
from imgmarker.coordinates import PixCoord, WorldCoord
import os
from math import nan, ceil
from astropy.wcs.utils import proj_plane_pixel_scales
from typing import TYPE_CHECKING, overload, Literal
import warnings

if TYPE_CHECKING:
    from imgmarker.image import Image 

class MarkLabel(QGraphicsProxyWidget):
    """Mark label and its attributes associated with a particular mark"""

    def __init__(self,mark:'Mark'):
        super().__init__()
        self.mark = mark
        self.lineedit = QLineEdit()
        self.lineedit.setReadOnly(True)
        f = self.lineedit.font()
        f.setPixelSize(int(self.mark.size))
        self.lineedit.setFont(f)

        # Using TabFocus because PyQt does not allow only focusing with left click
        self.setFocusPolicy(Qt.FocusPolicy.TabFocus)
        self.lineedit.setFocusPolicy(Qt.FocusPolicy.TabFocus)

        self.lineedit.setText(self.mark.text)
        self.lineedit.setStyleSheet(f"""background-color: rgba(0,0,0,0);
                                     border: none; 
                                     color: rgba{self.mark.color.getRgb()}""")
        
        self.lineedit.textChanged.connect(self.autoresize)
        self.setWidget(self.lineedit)
        self.autoresize()
        self.installEventFilter(self)
        self.setPos(self.mark.view_center+QPointF(self.mark.size/2,self.mark.size/2))

    def enter(self):
        self.setCursor(Qt.CursorShape.ArrowCursor)
        self.clearFocus()
        self.mark.text = self.lineedit.text()
        self.lineedit.setReadOnly(True)

    def focusInEvent(self, event):
        self.setCursor(Qt.CursorShape.IBeamCursor)
        self.lineedit.setReadOnly(False)
        return super().focusInEvent(event)
        
    def keyPressEvent(self, event):
        if (event.key() == Qt.Key.Key_Return): self.enter()
        else: return super().keyPressEvent(event)

    def eventFilter(self, source, event):
        if (event.type() == QEvent.Type.MouseButtonPress) or (event.type() == QEvent.Type.MouseButtonDblClick):
            if event.button() == Qt.MouseButton.LeftButton:
                # With TabFocusReason, tricks PyQt into doing proper focus events
                self.setFocus(Qt.FocusReason.TabFocusReason)
            return True
        return super().eventFilter(source,event)
        
    def autoresize(self):
        fm = self.lineedit.fontMetrics()
        w = fm.boundingRect(self.lineedit.text()).width()+fm.boundingRect('AA').width()
        h = fm.boundingRect('AA').height()
        self.lineedit.setFixedWidth(w)
        self.lineedit.setFixedHeight(h)

class Mark(QGraphicsPathItem):
    """Class for creating marks and associating label to mark"""

    @overload
    def __init__(self,x:int,y:int,
                 shape:str='ellipse',
                 image:'Image'=None,group:int=0,text:str=None,color:QColor=None,size_unit:Literal['px','arcsec']=None,size:float=None,
    ) -> None: ...
    @overload
    def __init__(self,ra:float=None,dec:float=None,
                 shape:str='ellipse',
                 image:'Image'=None,group:int=0,text:str=None,color:QColor=None,size_unit:Literal['px','arcsec']=None,size:float=None,
    ) -> None: ...
    def __init__(self,*args,**kwargs) -> None:
        # Set up some default values
        self.image = None
        self.g = 0
        self._shape = 'ellipse'
        self.size_unit = 'px'
        self.dst = os.path.join(config.SAVE_DIR,f'{config.USER}_marks.csv')
        self.label:MarkLabel = QGraphicsProxyWidget()

        if 'image' in kwargs: 
            self.image:'Image' = kwargs['image']

        if ('group' in kwargs) and (kwargs['group'] != 0):
            self.g:int = kwargs['group']
            self._color = config.GROUP_COLORS[self.g]

        if 'color' in kwargs:
            self._color = kwargs["color"]

        if 'text' in kwargs:
            self.text:str = kwargs['text']
        else:
            self.text = config.GROUP_NAMES[self.g]

        if 'shape' in kwargs:
            self._shape = kwargs['shape']
        
        if "size" in kwargs:
            if 'size_unit' in kwargs:
                self.size_unit = kwargs['size_unit']
            self._size_value = kwargs['size']

        if 'ra' in kwargs:
            self._wcs_center = WorldCoord(kwargs['ra'],kwargs['dec'])            
        else:
            self._center = PixCoord(*args)
        
        super().__init__()
        
        self.setFlag(self.GraphicsItemFlag.ItemIsSelectable)

    @property
    def color(self):
        if hasattr(self,'_color'):
            return self._color
        else:
            return config.DEFAULT_COLORS[self.dst]

    @property
    def size_value(self):
        if hasattr(self,'_size_value'):
            return self._size_value
        elif self.image != None:
            return ceil((self.image.width+self.image.height)/200)*2
        else:
            return 10
    
    @property
    def size(self):
        if self.size_unit == "arcsec":
            pixel_scale = proj_plane_pixel_scales(self.image.wcs)[0] * 3600
            return self.size_value / pixel_scale
        elif self.size_unit == "px":
            return self.size_value
        else:
            warnings.warn("Invalid size unit for catalog marks. Valid units: arcsec, px")
            return
        
    @property
    def center(self) -> PixCoord:
        if not hasattr(self,'_center'):
            return self.wcs_center.topix(self.image.wcs)
        else:
            return self._center
    
    @property
    def view_center(self):
        return QPointF(*self.center) + QPointF(0.5,0.5)

    @property
    def wcs_center(self) -> WorldCoord:
        if not hasattr(self,'_wcs_center'):
            if (self.image.wcs != None):
                return self.center.toworld(self.image.wcs)
            else: 
                return (nan, nan)
        else:
            return self._wcs_center

    def draw(self):
        args = (self.view_center.x()-self.size/2,
                self.view_center.y()-self.size/2,
                self.size,
                self.size)
        
        path = QPainterPath()

        if self._shape == 'ellipse':
            path.addEllipse(*args)
        elif self._shape == 'rect':
            path.addRect(*args)
        
        self.setPath(path)
        pen = QPen(self.color, # brush
                int(self.size/10), # width
                Qt.PenStyle.SolidLine, # style
                Qt.PenCapStyle.RoundCap, # cap
                Qt.PenJoinStyle.MiterJoin) # join
        self.setPen(pen)




"""
Copyright © 2025, UChicago Argonne, LLC

Full license found at _YOUR_INSTALLATION_DIRECTORY_/imgmarker/LICENSE
"""

"""This module contains code for the `Image` class and image manipulation."""

from imgmarker.gui.pyqt import QImage, QGraphicsPixmapItem, QGraphicsView, QGraphicsScene, QPixmap, Qt, QPointF, QApplication
from imgmarker.gui import Mark, MarkLabel
from io import StringIO
import os
from math import floor
import PIL.Image as pillow
from PIL.TiffTags import TAGS
from math import nan
import numpy as np
from typing import overload, Union, List, Set
from astropy.visualization import ZScaleInterval, MinMaxInterval, ManualInterval, LinearStretch, LogStretch
from . import fits
from .convolution import gaussian_filter
from astropy.wcs import WCS
from enum import Enum
from copy import deepcopy
import warnings

class Interval:
    ZSCALE = ZScaleInterval()
    MINMAX = MinMaxInterval()
    
class Stretch:
    LINEAR = LinearStretch()
    LOG = LogStretch()  

class Mode(Enum):
    RGB = 0
    RGBA = 1
    I8 = 2
    L = 2
    I16 = 3
    
    def __init__(self,value):
        self.format = {0: QImage.Format.Format_RGB888, 
                       1: QImage.Format.Format_RGBA8888,
                       2: QImage.Format.Format_Grayscale8,
                       3: QImage.Format.Format_Grayscale16}[value]
        self.iinfo:np.iinfo = {0: np.iinfo(np.uint8), 
                               1: np.iinfo(np.uint8),
                               2: np.iinfo(np.uint8),
                               3: np.iinfo(np.uint16)}[value]

FORMATS = ['TIFF','FITS','PNG','JPEG']

pillow.MAX_IMAGE_PIXELS = None # change this if we want to limit the image size

def pathtoformat(path:str):
    ext = path.split('.')[-1].casefold()
    if ext == 'png': return 'PNG'
    if ext in {'jpeg', 'jpg'}: return 'JPEG'
    if ext in {'tiff', 'tif'}: return 'TIFF'
    if ext in {'fit', 'fits'}: return 'FITS'

def align8to32(bytes: bytes, width: int, bits_per_pixel: str) -> bytes:
    """
    converts each scanline of data from 8 bit to 32 bit aligned. slightly modified from Pillow
    """

    # calculate bytes per line and the extra padding if needed
    bits_per_line = bits_per_pixel * width
    full_bytes_per_line, remaining_bits_per_line = divmod(bits_per_line, 8)
    bytes_per_line = full_bytes_per_line + (1 if remaining_bits_per_line else 0)

    extra_padding = -bytes_per_line % 4

    # already 32 bit aligned by luck
    if not extra_padding:
        return bytes

    new_data = [
        bytes[i * bytes_per_line : (i + 1) * bytes_per_line] + b"\x00" * extra_padding
        for i in range(len(bytes) // bytes_per_line)
    ]

    return b"".join(new_data)

def vibrance(c,mode:Mode):
    if mode == Mode.RGB:
        v = np.max(c, axis=2)
    elif mode == Mode.RGBA:
        v = np.max(c[:, :, :3], axis=2)
    else: 
        v = c
    return v

def read_wcs(f):
    """Reads WCS information from headers if available. Returns `astropy.wcs.WCS`."""
    try:
        if isinstance(f,fits.PrimaryHDU) or isinstance(f,fits.ImageHDU):
            
            if not 'CRPIX1' in f.header.keys(): return None
            else: 
                _header = f.header.copy()
                _header['NAXIS'] = 2
                return WCS(_header)
        else:
            meta_dict = {TAGS[key] : f.tag[key] for key in f.tag_v2}
            
            long_header_str = meta_dict['ImageDescription'][0]

            line_length = 80

            # Splitting the string into lines of 80 characters
            lines = [long_header_str[i:i+line_length] for i in range(0, len(long_header_str), line_length)]
            
            # Join the lines with newline characters to form a properly formatted header string
            corrected_header_str = "\n".join(lines)

            # Use an IO stream to mimic a file
            header_stream = StringIO(corrected_header_str)

            # Read the header using astropy.io.fits
            header = fits.Header.fromtextfile(header_stream)

            # Create a WCS object from the header
            return WCS(header)
    except: return None

class Image(QGraphicsPixmapItem):
    """
    Image class based on the PyQt QGraphicsPixmapItem.

    Attributes
    ----------
    path: str
        Path to the image.

    name: str
        File name.

    format: str
        Image format. Can be TIFF, FITS, PNG, or JPEG.

    frame: int
        Current frame of the image.

    n_frame: int
        Number of frames in the image.

    imagefile: `PIL.Image.ImageFile`
        Pillow imagefile object that allows loading of image data and image manipulation.

    width: int
        Image width.

    height: int
        Image height.

    wcs: `astropy.wcs.WCS` or None
        WCS solution.

    wcs_center: list[float]
        Center of the image in WCS coordinates.

    r: float
        Blur radius applied to the image.

    stretch: `BaseStretch`, default=`LinearStretch()`
        Stretch of the image brightness. Can be set with `Image.stretch = 'linear'` or `Image.stretch = 'log'`

    interval: `BaseInterval`, default=`ZScaleInterval()`
        Interval of the image brightness. Can be set with `Image.interval = 'zscale'` or `Image.stretch = 'min-max'`

    comment: str
        Image comment.

    categories: list[int]
        List containing the categories for this image.

    marks: list[imgmarker.mark.Mark]
        List of the marks in this image.

    dupe_marks: list[imgmarker.mark.Mark]
        List of marks made on a duplicate-showing of the same image. Kept separate for saving purposes.

    seen: bool
        Whether this image has been seen by the user or not.

    markfiles: list[str]
        List of paths to the markfiles that have been imported for this image
    """
    
    def __init__(self,path:str):
        """
        Parameters
        ----------
        path: str 
            Path to the image.
        """
        
        super().__init__(QPixmap())

        self.path = path
        self.name = path.split(os.sep)[-1]
        self.format = pathtoformat(path)
        self.incompatible = False
        if self.format in FORMATS:
            
            self.frame:int = 0
            metadata = self.read_metadata()
            if metadata != None:
                self.incompatible = False
                self.duplicate = False
                self.width = metadata['width']
                self.height = metadata['height']
                self.mode:Mode = metadata['mode']
                self.n_channels = metadata['n_channels'] 
                self.n_frames = metadata['n_frames']
                self.wcs:WCS = metadata['wcs']

                self.r:float = 0.0
                self.stretch = Stretch.LINEAR
                self.interval = Interval.MINMAX
                
                self.comment = 'None'
                self.categories:List[int] = []
                self.marks:List['Mark'] = []
                self.dupe_marks:List['Mark'] = []
                self.undone_marks:List['Mark'] = []
                self.seen:bool = False
                self.markfiles:Set[str] = set()
            else:
                self.incompatible = True

    def __deepcopy__(self, memo):
        id_self = id(self)
        _copy = memo.get(id_self)
        if _copy is None:
            _copy = type(self)(
                deepcopy(self.path, memo))
            memo[id_self] = _copy
        return _copy

    @property
    def scaling(self):
        return self.stretch + ManualInterval(*self.vlims)

    @property
    def v(self):
        return vibrance(self.array,self.mode)
    
    @property
    def vlims(self):
        v = vibrance(self._array,self.mode) 
        vlims = self.interval.get_limits(v)
        return vlims
    
    @property
    def wcs_center(self) -> list:
        try: return self.wcs.all_pix2world([[self.width/2, self.height/2]], 0)[0]
        except: return nan, nan

    def read(self) -> np.ndarray:
        if self.format == 'FITS':
            with fits.open(self.path) as f:
                data = f[self.frame].data
                if self.n_channels > 1:
                    data = np.dstack((data[0], data[1], data[2]))
                data = np.flipud(data)
                data = (2**self.mode.iinfo.bits - 1) * ((data - np.nanmin(data)) / (np.nanmax(data) - np.nanmin(data)))
        
        else:
            with pillow.open(self.path) as f:
                f.seek(self.frame)
                data = np.array(f)

        return data
    
    def read_metadata(self) -> dict:
        metadata = {}
        if self.format == 'FITS':
            with fits.open(self.path) as f:
                try:
                    metadata['width'] = f[self.frame].header['NAXIS2']
                    metadata['height'] = f[self.frame].header['NAXIS1']

                    # If more than 1 channel, set number of channels and force 8 bits per channel
                    
                    if 'NAXIS3' in f[self.frame].header.keys():
                        metadata['n_channels'] = abs(f[self.frame].header['NAXIS3'])
                        metadata['mode'] = Mode.RGB

                    # Otherwise set number of channels to 1, get the bit depth from header
                    else: 
                        metadata['n_channels'] = 1
                        if abs(f[self.frame].header['BITPIX']) == 8: metadata['mode'] = Mode.I8
                        else: metadata['mode'] = Mode.I16

                    metadata['n_frames'] = len(f)
                    metadata['wcs'] = read_wcs(f[self.frame])
                    
                except:
                    warnings.warn(f"File \"{self.name}\" is not compatible and will not be loaded. Skipping \"{self.name}\".")
                    self.incompatible = True
                    return None

        else:
            with pillow.open(self.path) as f: 
                f.seek(self.frame)
                metadata['width'] = f.width
                metadata['height'] = f.height
                metadata['mode'] = Mode[f.mode.replace(';','')]
                metadata['n_channels'] = len(f.getbands())
                try: metadata['n_frames'] = f.n_frames
                except: metadata['n_frames'] = 1
                metadata['wcs'] = read_wcs(f)
        
        return metadata
    
    def close(self):
        self._array = None
        self.array = None
        self.setPixmap(QPixmap())
    
    def seek(self,frame:int=0):
        """Switches to a new frame if it exists"""

        # Set frame
        frame = floor(frame)
        
        if frame > self.n_frames - 1: frame = 0
        elif frame < 0: frame = self.n_frames - 1

        self.frame = frame

        # Read data from the frame
        self._array = self.read()
        self.array = self._array
        self.width = self._array.shape[1]
        self.height = self._array.shape[0]
        
        # Rescale (and blur)
        with np.errstate(divide='ignore', invalid='ignore'):
            self.rescale()

    def toqimage(self,array:np.ndarray) -> QImage:
        width, height  = array.shape[1], array.shape[0]
        data = array.tobytes()

        if self.mode in {Mode.I16, Mode.I8, Mode.L}: 
            data = align8to32(data,width,self.mode.iinfo.bits)

        if array.ndim == 3:
            n = array.shape[2]
            qim = QImage(data,width,height,n*width,self.mode.format)
        else:
            qim = QImage(data,width,height,self.mode.format)

        return qim

    def topixmap(self,array:np.ndarray) -> QPixmap:
        """Creates a QPixmap from an array."""
        
        qimage = self.toqimage(array)
        pixmap = QPixmap.fromImage(qimage)

        return pixmap
    
    def rescale(self):
        out = self._array.astype(np.float64)

        if (self.mode == Mode.RGB) or (self.mode == Mode.RGBA):
            # Calculate scale factor
            v = self.v
            scale = self.mode.iinfo.max*self.scaling(v)/v

            # Apply scale factor
            out[:, :, 0] *= scale
            out[:, :, 1] *= scale
            out[:, :, 2] *= scale

            # Truncate values greater than the max pixel value for this mode
            out = np.minimum(self.mode.iinfo.max,out)

        else:
            out = self.mode.iinfo.max*self.scaling(out)

        self.array = out.astype(self.mode.iinfo.dtype)
        self.blur()
    
    @overload
    def blur(self) -> None: 
        """Applies the blur to the image"""
    @overload
    def blur(self,value) -> None:
        """Applies the blur to the image"""
    def blur(self,*args):
        if len(args) > 0: 
            value = args[0]
            if callable(value): r = value()
            else: r = value
            self.r = floor(r)/2

        nanmask = ~np.isfinite(self.array)
        if True in nanmask:
            # Create auxillary array
            _out = self.array.copy()
            _out[nanmask] = 0
            out = gaussian_filter(_out,self.r)

            # Calculate weights
            _w = np.ones_like(self.array)
            _w[nanmask] = 0
            w = gaussian_filter(_w,self.r)

            # Apply weights, add nans back in
            with np.errstate(divide='ignore', invalid='ignore'): 
                out = out/w
                out[nanmask] = np.nan
        else:
            out = gaussian_filter(self.array,self.r)

        self.setPixmap(self.topixmap(out.astype(self.mode.iinfo.dtype)))


class ImageScene(QGraphicsScene):
    """A class in which images and marks are stored."""
    def __init__(self,image:Image):
        super().__init__()
        self.image = image

        self.setBackgroundBrush(Qt.GlobalColor.black)
        self.addItem(self.image)
        self.setSceneRect(-4*self.image.width,-4*self.image.height,9*self.image.width,9*self.image.height)

    def mousePressEvent(self, event):
        modifiers = QApplication.keyboardModifiers()
        leftbutton = event.button() == Qt.MouseButton.LeftButton
        shift = modifiers == Qt.KeyboardModifier.ShiftModifier
        if not (leftbutton and shift):
            return super().mousePressEvent(event)
        
    

    def update_image(self,image:Image):
        """Updates the current image with a new image."""
        # Remove items
        for item in self.items(): self.removeItem(item)

        # Update the pixmap
        self.image = image
        self.addItem(self.image)
        self.setSceneRect(-4*self.image.width,-4*self.image.height,9*self.image.width,9*self.image.height)

    @overload
    def mark(self,x:float,y:float,shape='ellipse',text:Union[int,str]=0) -> 'Mark': ...
    @overload
    def mark(self,ra:float=None,dec:float=None,shape='ellipse',text:Union[int,str]=0) -> 'Mark': ...
    @overload
    def mark(self,mark:'Mark') -> 'Mark': ... 

    def mark(self,*args,**kwargs) -> 'Mark':
        """Creates a mark object and adds it to the image scene and returns the mark."""

        if len(args) == 1: mark = args[0]
        else: mark = Mark(*args,image=self.image,**kwargs)
        mark.draw()
        mark.label = MarkLabel(mark)
        self.addItem(mark)
        self.addItem(mark.label)
        return mark
    
    def rmmark(self,mark:'Mark') -> None:
        """Removes the specified mark from the image scene."""

        self.removeItem(mark)
        self.removeItem(mark.label)

class ImageView(QGraphicsView):
    """A class in which the image scene is stored."""
    def __init__(self,scene:ImageScene):
        super().__init__(scene)

        self.zoom_level = 1
        self.cursor_focus = False

        self.move(0, 0)
        self.setTransformationAnchor(self.ViewportAnchor(1))

        # Mouse tracking
        self.setMouseTracking(True)

        # Disable scrollbars
        self.verticalScrollBar().blockSignals(True)
        self.setVerticalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
        self.horizontalScrollBar().blockSignals(True)
        self.setHorizontalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOff)

        #Install event filter for zooming
        self.viewport().installEventFilter(self)

    def eventFilter(self, a0, a1):
        """
        Performs operations based on the event source and type.

        Parameters
        ----------
        source: `QObject` object
            Source of the event
        event: `QEvent` object
            Event

        Returns
        ----------
        True if the event triggered an some operation.
        """

        if (a0 == self.viewport()) and (a1.type() == 31):
            x = a1.angleDelta().y()/120
            self.zoom(1.2**(-x))
            return True

        return super().eventFilter(a0, a1)        
    
    def scene(self) -> ImageScene:
        """Returns the associated image scene."""
        return super().scene()
    
    def mouse_pix_pos(self,correction:bool=True):
        """
        Gets mouse image coordinates.

        Returns
        ----------
        pix_pos: `QPoint`
            position of mouse in the image.
        """

        view_pos = self.mapFromGlobal(self.cursor().pos())
        scene_pos = self.mapToScene(view_pos)

        # Get the pixel coordinates (including padding; half-pixel offset required)
        pix_pos = self.scene().image.mapFromScene(scene_pos)

        # Correct half-pixel error
        if correction: pix_pos -= QPointF(0.5,0.5)
        
        return pix_pos.toPoint()
    
    def mouse_pos(self):
        """
        Gets mouse position.

        Returns
        ----------
        view_pos: `QPoint`
            position of mouse in the image view.
        """

        return self.mapFromGlobal(self.cursor().pos())
    
    def zoom(self,scale:float,mode:str='mouse'):
        """
        Zoom in on the image.

        Parameters
        ----------
        scale: str
            Scale of the zoom. Greater than 1 means zooming in, less than 1 means zooming out
        mode: str, optional
            Zoom mode. To zoom from the center of the viewport, use mode='viewport'. To zoom from the mouse
            cursor location, use mode='mouse'. Defaults to 'mouse'.
        
        Returns
        ----------
        None
        """

        if self.zoom_level*scale > 1/3:
            self.zoom_level *= scale
            if mode == 'viewport': center = self.viewport().rect().center()
            if mode == 'mouse': center = self.mouse_pos()

            transform = self.transform()
            center = self.mapToScene(center)
            transform.translate(center.x(), center.y())
            transform.scale(scale, scale)
            transform.translate(-center.x(), -center.y())
            self.setTransform(transform)

    def zoomfit(self):
        """Fit the image view in the viewport."""

        self.fitInView(self.sceneRect(), Qt.AspectRatioMode.KeepAspectRatio)
        self.zoom(scale=9,mode='viewport')
        self.zoom_level = 1

    def center_cursor(self):
        """Center on the cursor."""

        center = self.viewport().rect().center()
        scene_center = self.mapToScene(center)
        pix_pos = self.mouse_pix_pos(correction=False)

        delta = scene_center.toPoint() - pix_pos
        self.translate(delta.x(),delta.y())
        
        if self.cursor_focus:
            global_center = self.mapToGlobal(center)
            self.cursor().setPos(global_center)
    





"""
Copyright © 2025, UChicago Argonne, LLC

Full license found at _YOUR_INSTALLATION_DIRECTORY_/imgmarker/LICENSE
"""

from PyQt6.QtWidgets import (
    QApplication, QMainWindow, QPushButton, QLabel, 
    QScrollArea, QGraphicsView, QVBoxLayout, QWidget, 
    QHBoxLayout, QLineEdit, QInputDialog, QCheckBox, 
    QSlider, QLineEdit, QFileDialog, QFrame, 
    QDialog, QSizePolicy, QGraphicsPathItem, QGraphicsProxyWidget,
    QLineEdit, QGraphicsScene, QGraphicsPixmapItem, QSpinBox,
    QMessageBox, QTableWidget, QTableWidgetItem, QHeaderView,
    QMenu, QColorDialog, QWidgetAction,
)

from PyQt6.QtGui import (
    QIcon, QFont, QClipboard, QAction,
    QPen, QColor, QPixmap, QPainter, QPainterPathStroker,
    QPainterPath, QImage, QShortcut, QDesktopServices,
    QBrush
)
    
from PyQt6.QtCore import (
    Qt, QPoint, QKeyCombination, QPointF, 
    QEvent, QUrl, PYQT_VERSION_STR
)


"""
Copyright © 2025, UChicago Argonne, LLC

Full license found at _YOUR_INSTALLATION_DIRECTORY_/imgmarker/LICENSE
"""

"""Image Marker's I/O module containing functions for reading and saving data."""

import os
import numpy as np
from imgmarker.gui import Mark
from imgmarker import image, config
import glob as _glob
from math import nan, isnan
from typing import Tuple, List
import csv
import datetime as dt

class MarkFile:
    VALID_FIELDNAMES = [
        'date','image','group',
        'label','x','y',
        'ra','dec','size'
        'size(px)','size(arcsec)'
    ]
    
    def __init__(self,path:str):
        self.path = path

        if path not in config.DEFAULT_COLORS:
            config.DEFAULT_COLORS[path] = config.GROUP_COLORS[0]

        valid, err = self.isvalid(return_err=True)
        if not valid:
            raise err
    
    def __eq__(self, value):
        if hasattr(value,'path'):
            return self.path == value.path
        else:
            return self.path == value
        
    def isvalid(self,return_err=False):
        valid = True
        err = None

        if os.path.exists(self.path):
            with open(self.path,'r') as f:
                delimiter = '|' if '|' in f.readline() else ','
                f.seek(0)
                reader = csv.DictReader(f,delimiter=delimiter)
                for fieldname in reader.fieldnames:
                    if fieldname.strip().lower() not in MarkFile.VALID_FIELDNAMES:
                        valid = False
                        err = KeyError(f'Field name "{fieldname}" in file "{self.path.split(os.sep)[-1]}" is not a valid field name.')

        if return_err:
            return valid, err
        else:
            return valid
     
    def read(self,images:List[image.Image]) -> Tuple[List[image.Image],List[Mark]]:
        """
        Takes data from marks.csv and images.csv and from them returns a list of `imgmarker.image.Image`
        objects.

        Returns
        ----------
        images: list[`imgmarker.image.Image`]
        """

        imageless = []
        
        # Get list of marks for each image
        if os.path.exists(self.path):
            with open(self.path,'r') as f:
                delimiter = '|' if '|' in f.readline() else ','
                f.seek(0)
                reader = csv.DictReader(f,delimiter=delimiter)

                for row in reader:
                    for fieldname in reader.fieldnames:
                        row[fieldname.strip().lower()] = row.pop(fieldname).strip()

                    # Default values
                    name = 'None'
                    group = config.GROUP_NAMES.index('None')
                    shape = 'rect'
                    label = 'None'
                    x,y = nan,nan
                    ra,dec = nan,nan
                    size = None
                    size_unit = None

                    # Values from row
                    if 'date' in row: date = row['date']

                    if 'image' in row: name = row['image']

                    if 'group' in row: 
                        group = config.GROUP_NAMES.index(row['group'])
                        shape = config.GROUP_SHAPES[group]
                        
                    if 'label' in row: label = row['label']

                    if 'x' in row: x = float(row['x'])
                    if 'y' in row: y = float(row['y'])

                    if 'ra' in row: ra = float(row['ra'])
                    if 'dec' in row: dec = float(row['dec'])

                    if 'size(arcsec)' in row: 
                        size = float(row['size(arcsec)'])
                        size_unit = 'arcsec'
                    
                    if 'size(px)' in row:
                        size = float(row['size(px)'])
                        size_unit = 'px'

                    if 'size' in row:
                        size = float(row['size'])
                        size_unit = 'px'

                    if name != 'None':
                        for img in images:
                            if (name == img.name) and (not isnan(float(x))) and (not isnan(float(y))):
                                args = (float(x),float(y))
                                kwargs = {'image': img, 'group': group, 'shape': shape}

                                if label != 'None': 
                                    kwargs['text'] = label

                                if size != None: 
                                    kwargs['size'] = size
                                    kwargs['size_unit'] = size_unit

                                mark = Mark(*args, **kwargs)
                                mark.dst = self.path
                                img.marks.append(mark)

                    else:
                        if isnan(float(ra)) and isnan(float(dec)):
                            args = (float(x),float(y))
                            kwargs = {'image': None, 'group': group, 'shape': shape}
                        else:
                            args = ()
                            kwargs = {'image': None,'group': group, 'shape': shape, 'ra': ra, 'dec': dec}

                        if label != 'None': 
                            kwargs['text'] = label

                        if size != None: 
                            kwargs['size'] = size
                            kwargs['size_unit'] = size_unit
                        
                        mark = Mark(*args, **kwargs)
                        mark.dst = self.path
                        imageless.append(mark)

        return images, imageless
    
    def save(self,images:List[image.Image],imageless_marks:List[Mark]) -> None:
        """
        Saves mark data.

        Parameters
        ----------

        images: list[`imgmarker.image.Image`]
            A list of Image objects for each image from the specified image directory.

        Returns
        ----------
        None
        """

        # Will organize output rows into dictionary of the path to save to

        date = dt.datetime.now(dt.timezone.utc).date().isoformat()
        rows = []

        for img in images:
            if img.seen:
                if img.duplicate == True:
                    marks = img.dupe_marks
                else:
                    marks = img.marks

                name = img.name

                if not marks: mark_list = [None]
                else: mark_list = marks.copy()
                
                for mark in mark_list:
                    
                    row = {}

                    if (mark != None) and (mark.dst == self.path):
                        group_name = config.GROUP_NAMES[mark.g]

                        if mark.text == group_name: 
                            label = 'None'
                        else: 
                            label = mark.text

                        if (img.duplicate == True) and (mark in img.dupe_marks):
                            if (mark.text == group_name):
                                label = "DUPLICATE"
                            else:
                                label = f"{mark.text}, DUPLICATE"

                        try: x, y = mark.center
                        except: x, y = nan, nan
                        
                        try: ra, dec = mark.wcs_center
                        except: ra, dec = nan, nan

                        row = {
                            'date': str(date),
                            'image': str(name),
                            'group': str(group_name),
                            'label': str(label),
                            'x': str(x),
                            'y': str(y),
                            'RA': str(ra),
                            'DEC': str(dec)
                        }
                        
                        rows.append(row)

                    # Row entries if the seen image has no marks (user looked at image without placing)
                    else:
                        row = {
                            'date': str(date),
                            'image': str(name),
                            'group': 'None',
                            'label': 'None',
                            'x': nan,
                            'y': nan,
                            'RA': nan,
                            'DEC': nan
                        }
                    
                        rows.append(row)

        for mark in imageless_marks:
            row = {}
            if mark.dst == self.path:
                group_name = config.GROUP_NAMES[mark.g]

                if mark.text == group_name: 
                    label = 'None'
                else: 
                    label = mark.text

                name = 'None'

                try: x, y = mark.center.x(), mark.center.y()
                except: x, y = nan, nan
                
                try: ra, dec = mark.wcs_center
                except: ra, dec = nan, nan

                row = {
                    'date': str(date),
                    'image': str(name),
                    'group': str(group_name),
                    'label': str(label),
                    'x': str(x),
                    'y': str(y),
                    'RA': str(ra),
                    'DEC': str(dec)
                }
                    
                rows.append(row)
        
        # Write lines if there are lines to print
        with open(self.path, 'w', newline='') as f:
            writer = csv.DictWriter(f, fieldnames=rows[0].keys())
            writer.writeheader()
            for row in rows:
                writer.writerow(row)

class ImagesFile:
    def __init__(self):
        self.path = os.path.join(config.SAVE_DIR,f'{config.USER}_images.csv')
    
    def __eq__(self, value):
        if hasattr(value,'path'):
            return self.path == value.path
        else:
            return self.path == value
        
    def read(self) -> Tuple[List[image.Image],List[Mark]]:
        """
        Takes data from marks.csv and images.csv and from them returns a list of `imgmarker.image.Image`
        objects.

        Returns
        ----------
        images: list[`imgmarker.image.Image`]
        """

        images:List[image.Image] = []
        
        # Get list of images from images.csv
        if os.path.exists(self.path):
            with open(self.path,'r') as f:
                delimiter = '|' if '|' in f.readline() else ','
                f.seek(0)
                reader = csv.DictReader(f,delimiter=delimiter)
                
                for row in reader:
                    keys = row.copy().keys()
                    for key in keys: row[key.strip().lower()] = row.pop(key).strip()

                    ra,dec = float(row['ra']), float(row['dec'])
                    date,name,categories,comment = row['date'], row['image'], row['categories'], row['comment']
                    categories = categories.split('+')
                    categories = [config.CATEGORY_NAMES.index(cat) for cat in categories if cat != 'None']
                    categories.sort()

                    img = image.Image(os.path.join(config.IMAGE_DIR,name))
                    img.comment = comment
                    img.categories = categories
                    img.seen = True
                    images.append(img)

        return images

    def save(self, images:List['image.Image']):
        """
        Saves image data.

        Parameters
        ----------
        images: list[`imgmarker.image.Image`]
            A list of Image objects for each image from the specified image directory.

        Returns
        ----------
        None
        """

        date = dt.datetime.now(dt.timezone.utc).date().isoformat()
        image_rows:list[dict] = []
    
        for img in images:
            if img.seen:
                name = img.name
                comment = img.comment

                category_list = img.categories
                category_list.sort()
                if (len(category_list) != 0):
                    categories = '+'.join([config.CATEGORY_NAMES[i] for i in category_list])
                else: categories = 'None'

                image_rows.append({'date': str(date),
                                    'image': str(name),
                                    'RA': str(img.wcs_center[0]),
                                    'DEC': str(img.wcs_center[1]),
                                    'categories': str(categories),
                                    'comment': str(comment)})

        with open(self.path, 'w') as f:
            writer = csv.DictWriter(f, fieldnames=image_rows[0].keys())
            writer.writeheader()
            for row in image_rows:
                writer.writerow(row)

class FavoritesFile:
    def __init__(self):
        self.path = os.path.join(config.SAVE_DIR,f'{config.USER}_favorites.csv')
    
    def __eq__(self, value):
        if hasattr(value,'path'):
            return self.path == value.path
        else:
            return self.path == value
        
    def read(self) -> List[str]:
        """
        Takes data from favorites.csv and from them returns a list of image file names
        with full directory.

        Returns
        ----------
        favorites: list[`str`]
        """

        favorites:List[str] = []
        
        # Get list of images from favorites.csv
        if os.path.exists(self.path):
            with open(self.path,'r') as f:
                delimiter = '|' if '|' in f.readline() else ','
                f.seek(0)
                reader = csv.DictReader(f,delimiter=delimiter)
                
                for row in reader:
                    keys = row.copy().keys()
                    for key in keys: row[key.strip().lower()] = row.pop(key).strip()
                    date,name,categories,comment = row['date'], row['image'], row['categories'], row['comment']

                    favorites.append(name)
        return favorites

    def save(self, favorites:List[str], images:List['image.Image']):
        """
        Saves favorites data.

        Parameters
        ----------
        favorites: list[str]
            A list of strings of each image file name from the specified image directory.

        images: list['image.Image']
            A list of image objects.
            
        Returns
        ----------
        None
        """
        # fav_out_path = os.path.join(config.SAVE_DIR, f'{config.USER}_favorites.csv')
        date = dt.datetime.now(dt.timezone.utc).date().isoformat()
        image_rows:list[dict] = []
    
        favorited = [favorite for favorite in images if favorite.name in favorites]

        for img in favorited:
            if img.seen:
                name = img.name
                comment = img.comment

                category_list = img.categories
                category_list.sort()
                if (len(category_list) != 0):
                    categories = '+'.join([config.CATEGORY_NAMES[i] for i in category_list])
                else: categories = 'None'

                image_rows.append({'date': str(date),
                                    'image': str(name),
                                    'RA': str(img.wcs_center[0]),
                                    'DEC': str(img.wcs_center[1]),
                                    'categories': str(categories),
                                    'comment': str(comment)})
        if len(favorited) > 0:
            with open(self.path, 'w') as f:
                writer = csv.DictWriter(f, fieldnames=image_rows[0].keys())
                writer.writeheader()
                for row in image_rows:
                    writer.writerow(row)
        else:
            with open(self.path, 'w') as f:
                f.write('')
        
def markpaths() -> List[str]:
    _paths = [os.path.join(config.SAVE_DIR,f'{config.USER}_marks.csv')]
    import_dir = os.path.join(config.SAVE_DIR,'imports')

    if not os.path.exists(import_dir):
        os.makedirs(import_dir)
    
    _paths += _glob.glob(os.path.join(import_dir,'*'))
    paths = []

    for path in _paths:
        try: paths.append(MarkFile(path).path)
        except: pass

    return paths

def glob(edited_images:List[image.Image]=[]) -> Tuple[List[image.Image],int]:
    """
    Globs in IMAGE_DIR, using edited_images to sort, with edited_images in order at the beginning of the list
    and the remaining unedited images in randomized order at the end of the list.

    Parameters
    ----------
    edited_images: list['imgmarker.image.Image']
        A list of Image objects containing the loaded-in information for each edited image.

    Returns
    ----------
    images: list['imgmarker.image.Image']
        A list of Image objects with the ordered edited images first and randomized unedited
        images added afterwards.
    
    idx: int
        The index to start at to not show already-edited images from a previous save.
    """

    # Find all images in image directory
    paths = sorted(_glob.glob(os.path.join(config.IMAGE_DIR, '*.*')))
    paths = [fp for fp in paths if image.pathtoformat(fp) in image.FORMATS]

    # Get list of paths to images if they are in the dictionary (have been edited)
    edited_paths = [os.path.join(config.IMAGE_DIR,img.name) for img in edited_images]
    unedited_paths = [fp for fp in paths if fp not in edited_paths]

    if config.RANDOMIZE_ORDER:
        # Shuffle the remaining unedited images
        rng = np.random.default_rng()
        rng.shuffle(unedited_paths)

    # Put edited images at the beginning, unedited images at front
    images = edited_images + [image.Image(fp) for fp in unedited_paths]
    for img in images:
        if img.incompatible == True:
            images.remove(img)

    idx = min(len(edited_images),len(paths)-1)

    return images, idx


"""
Copyright © 2025, UChicago Argonne, LLC

Full license found at _YOUR_INSTALLATION_DIRECTORY_/imgmarker/LICENSE
"""

from imgmarker.gui.pyqt import QApplication, QIcon
from imgmarker.gui.window import MainWindow, _open_save
from imgmarker import config, ICON
import sys

def run():
    app = QApplication(sys.argv)
    app.setWindowIcon(QIcon(ICON))
    
    config.SAVE_DIR = _open_save()
    config.IMAGE_DIR, config.GROUP_NAMES, config.CATEGORY_NAMES, config.GROUP_MAX, config.RANDOMIZE_ORDER = config.read()

    window = MainWindow()
    window.show()
    window.image_view.zoomfit()
    sys.exit(app.exec())

if __name__ == '__main__': 
    run()


"""
Copyright © 2025, UChicago Argonne, LLC

Full license found at _YOUR_INSTALLATION_DIRECTORY_/imgmarker/LICENSE
"""

"""Configuration options and functions for interacting with the configuration file are contained within this module."""

import os
from getpass import getuser
from typing import Tuple, List
from imgmarker.gui.pyqt import Qt, QColor

SAVE_DIR = os.path.expanduser('~')
USER = getuser()
IMAGE_DIR = None
GROUP_NAMES = ['None','1','2','3','4','5','6','7','8','9']
GROUP_SHAPES = ['rect'] + ['ellipse']*9

GROUP_COLORS = [
    QColor(255,255,255),
    QColor(255,0,0),
    QColor(255,128,0),
    QColor(255,255,0),
    QColor(0,255,0),
    QColor(0,255,255),
    QColor(0,128,128),
    QColor(0,0,255),
    QColor(128,0,255),
    QColor(255,0,255)
]

DEFAULT_COLORS:dict[str,QColor] = {}

CATEGORY_NAMES = ['None','1','2','3','4','5']
GROUP_MAX = ['None','None','None','None','None','None','None','None','None']
RANDOMIZE_ORDER = False

MARK_KEYBINDS = {
    1: {Qt.MouseButton.LeftButton,Qt.Key.Key_1}, 
    2: {Qt.Key.Key_2}, 
    3: {Qt.Key.Key_3}, 
    4: {Qt.Key.Key_4}, 
    5: {Qt.Key.Key_5}, 
    6: {Qt.Key.Key_6}, 
    7: {Qt.Key.Key_7}, 
    8: {Qt.Key.Key_8}, 
    9: {Qt.Key.Key_9}
}
        
def path():
    return os.path.join(SAVE_DIR,f'{USER}_config.txt')
    
def read() -> Tuple[str,List[str],List[str],List[str],List[int]]:
    """
    Reads in each line from {username}_config.txt. If there is no configuration file,
    a default configuration file will be created using the required text
    format.

    Returns
    ----------
    image_dir: str
        Directory containing desired image files.

    group_names: list[str]
        A list of containing labels for each mark button.

    category_names: list[str]
        A list containing labels for each image category.

    group_max: list[int]
        A list containing the maximum allowed number of marks for each group.
    """

    # If the config doesn't exist, create one
    if not os.path.exists(path()):
        with open(path(),'w') as config:
            image_dir = IMAGE_DIR
            group_names = GROUP_NAMES
            category_names = CATEGORY_NAMES
            group_max = GROUP_MAX
            randomize_order = RANDOMIZE_ORDER

            config.write(f'image_dir = {image_dir}\n')
            config.write(f"groups = {','.join(group_names)}\n")
            config.write(f"categories = {','.join(category_names)}\n")
            config.write(f"group_max = {','.join(group_max)}\n")
            config.write(f'randomize_order = {randomize_order}')  

    else:
        for l in open(path()):
            var, val = [i.strip() for i in l.replace('\n','').split('=')]

            if var == 'image_dir':
                if val == './': image_dir = os.getcwd()
                else: image_dir = val
                image_dir =  os.path.join(image_dir,'')

            if var == 'groups':
                group_names = []
                group_names_temp = val.split(',')
                for group_name in group_names_temp:
                    group_names.append(group_name.strip())
                group_names.insert(0, 'None')

            if var == 'categories':
                category_names = []
                category_names_temp = val.split(',')
                for category_name in category_names_temp:
                    category_names.append(category_name.strip())
                category_names.insert(0, 'None')
            
            if var == 'group_max':
                group_max = []
                group_max_temp = val.split(',')
                for group_max_val in group_max_temp:
                    group_max.append(group_max_val.strip())

            if var == 'randomize_order':
                randomize_order = val == 'True'

    return image_dir, group_names, category_names, group_max, randomize_order

def update() -> None:
    """Updates any of the config variables with the corresponding parameter."""
    
    with open(path(),'w') as config:
        config.write(f'image_dir = {IMAGE_DIR}\n')
        config.write(f"groups = {','.join(GROUP_NAMES[1:])}\n")
        config.write(f"categories = {','.join(CATEGORY_NAMES[1:])}\n")
        config.write(f"group_max = {','.join(GROUP_MAX)}\n")
        config.write(f'randomize_order = {RANDOMIZE_ORDER}')


"""
Copyright © 2025, UChicago Argonne, LLC

Full license found at _YOUR_INSTALLATION_DIRECTORY_/imgmarker/LICENSE
"""

import numpy as np
from scipy.ndimage import uniform_filter

def sigma_to_size(sigma, n:int):
    """
    Calculates the equivalent box filter sizes from a Gaussian filter radius.

    Algorithm taken from P. Kovesi (2010).

    Parameters
    ----------
    sigma: 
        Radius of the Gaussian filter.
    n: int
        Number of equivalent box filters.
    
    Returns
    ----------
    sizes: list
        Equivalent box filter sizes.

    """
    w = np.sqrt((12*sigma**2/n)+1)
    
    wl = np.floor(w)
    if wl % 2 == 0: wl -= 1
    wu = wl+2
				
    m = int((12*sigma**2 - n*wl*wl - 4*n*wl - 3*n)/(-4*wl - 4))
	
    sizes = [wl]*m + [wu]*(n-m) 

    return sizes

def gaussian_filter(array:np.ndarray, sigma, n:int=3) -> np.ndarray:
    """
    Applies an approximate gaussian filter as a series of `n` box filters.
    
    Parameters
    ----------
    array: `numpy.ndarray`
        Array to be filtered.
    sigma: 
        Radius of the Gaussian filter.
    n: int, optional
        Number of equivalent box filters. Default is 3.
    
    Returns
    ----------
    out: `numpy.ndarray`
        Filtered array.
    """

    out = array.copy()
    if sigma > 0:
        sizes = sigma_to_size(sigma, n)
        for s in sizes:
            if out.ndim == 3:
                for i in range(out.shape[2]):
                    out[:, :, i] = uniform_filter(out[:, :, i], s)
            else:
                out = uniform_filter(out,s)
    return out



"""
Copyright © 2025, UChicago Argonne, LLC

Full license found at _YOUR_INSTALLATION_DIRECTORY_/imgmarker/LICENSE
"""

import os
import pytest
from pytestqt.qtbot import QtBot
import numpy as np
from getpass import getuser
import datetime as dt
from imgmarker.gui.window import MainWindow
from imgmarker import gui, config, io, image

test_save_dir = os.path.abspath("./tests/test_save/")
test_images_dir = os.path.abspath("./tests/test_images/")
test_catalog_dir_txt = os.path.abspath("./tests/TEST_catalog.txt")
test_catalog_dir_csv = os.path.abspath("./tests/TEST_catalog.csv")

USER = getuser()

if os.path.exists(test_save_dir):
    os.remove(test_save_dir + f"{USER}_config.txt")
    os.remove(test_save_dir + f"{USER}_marks.csv")
    os.remove(test_save_dir + f"{USER}_images.csv")
    os.rmdir(test_save_dir)
    os.mkdir(test_save_dir)
else:
    os.mkdir(test_save_dir)

@pytest.fixture
def app(qtbot:QtBot):
    config.SAVE_DIR = test_save_dir
    config.IMAGE_DIR = test_images_dir
    config.USER = USER
    test_app = MainWindow()
    qtbot.addWidget(test_app)
    return test_app

def test_load_images(app:MainWindow, qtbot:QtBot):

    assert len(app.images) == 3

def test_image_shown(app:MainWindow, qtbot:QtBot):
    test_load_images(app, qtbot)
    image = app.image_scene.image
    assert image in app.image_scene.items()

def test_import_markfile(app:MainWindow, qtbot:QtBot):
    app.import_markfile(src=test_catalog_dir_csv)

    assert len(io.markpaths()) == 2
    assert len(app.image_scene.items()) == 3

def test_place_mark(app:MainWindow, qtbot):
    nitems_init = len(app.image_scene.items())
    nmarks_init = len(app.image.marks)
        
    app.mark(group=1, test=True)

    assert len(app.image_scene.items()) == nitems_init + 2
    assert len(app.image.marks) == nmarks_init + 1

def test_mark_limit(app:MainWindow, qtbot:QtBot):
    app.del_usermarks(mode='all')

    nitems_init = len(app.image_scene.items())
    nmarks_init = len(app.image.marks)

    config.GROUP_MAX[0] = 1
    config.GROUP_MAX[1] = 2

    app.mark(group=1, test=True)
    app.mark(group=1, test=True)
    app.mark(group=2, test=True)
    app.mark(group=2, test=True)
    app.mark(group=2, test=True)

    assert len(app.image_scene.items()) == nitems_init + 6
    assert len(app.image.marks) == nmarks_init + 3

def test_mark_delete(app:MainWindow, qtbot:QtBot):
    app.del_usermarks(mode='all')

    nitems_init = len(app.image_scene.items())
    nmarks_init = len(app.image.marks)

    app.mark(group=1, test=True)
    app.mark(group=2, test=True)
    app.mark(group=3, test=True)
    app.del_usermarks(mode='all')

    assert len(app.image_scene.items()) == nitems_init
    assert len(app.image.marks) == nmarks_init

def test_catalog_delete(app:MainWindow, qtbot:QtBot):

    nitems_init = [len(app.image_scene.items())]
    app.shift(+1)
    nitems_init.append(len(app.image_scene.items())) 
    app.shift(+1)
    nitems_init.append(len(app.image_scene.items()))
    app.shift(+1)

    app.import_markfile(src=test_catalog_dir_txt)
    assert len(app.image_scene.items()) == nitems_init[0] + 2

    app.shift(+1)
    assert len(app.image_scene.items()) == nitems_init[1] + 2
    
    app.shift(+1)
    assert len(app.image_scene.items()) == nitems_init[2] + 2

    app.del_markfile(os.path.join(config.SAVE_DIR,'imports',test_catalog_dir_txt.split(os.sep)[-1]))
    assert len(app.image_scene.items()) == nitems_init[2]

    app.shift(+1)
    assert len(app.image_scene.items()) == nitems_init[0]

    app.shift(+1)
    assert len(app.image_scene.items()) == nitems_init[1]

def test_frame_seek(app:MainWindow, qtbot:QtBot):
    first_frame_array = app.image.array
    app.image.seek(1)
    second_frame_array = app.image.array

    assert not np.all(first_frame_array == second_frame_array)

def test_save_mark(app:MainWindow, qtbot:QtBot):
    app.mark(group=1, test=True)
    date = 0
    name = 0
    group = 0
    label = 0
    x = 0
    y = 0
    ra = 0
    dec = 0
    line0 = True
    for line in open(os.path.join(config.SAVE_DIR,f'{config.USER}_marks.csv')):
        if line0: line0 = False
        else:
            date,name,group,label,x,y,ra,dec = [i.strip() for i in line.replace('\n','').split(',')]

    assert date == dt.datetime.now(dt.timezone.utc).date().isoformat()
    assert name == app.image.name
    assert group == config.GROUP_NAMES[1]
    assert label == "None"
    assert x == str(float(app.image.width/2))
    assert y == str(float(app.image.height/2))
    assert ra == str(app.image.marks[0].wcs_center[0])
    assert dec == str(app.image.marks[0].wcs_center[1])

def test_delete_save_mark(app:MainWindow, qtbot:QtBot):
    app.mark(group=1, test=True)
    date = 0
    name = 0
    group = 0
    label = 0
    x = 0
    y = 0
    ra = 0
    dec = 0
    line0 = True
    for line in open(os.path.join(config.SAVE_DIR,f'{config.USER}_marks.csv')):
        if line0: line0 = False
        else:
            date,name,group,label,x,y,ra,dec = [i.strip() for i in line.replace('\n','').split(',')]

    assert date == dt.datetime.now(dt.timezone.utc).date().isoformat()
    assert name == app.image.name
    assert group == config.GROUP_NAMES[1]
    assert label == "None"
    assert x == str(float(app.image.width/2))
    assert y == str(float(app.image.height/2))
    assert ra == str(app.image.marks[0].wcs_center[0])
    assert dec == str(app.image.marks[0].wcs_center[1])

    app.del_usermarks(mode='all')
    line0 = True
    for line in open(os.path.join(config.SAVE_DIR,f'{config.USER}_marks.csv')):
        if line0: line0 = False
        else:
            date,name,group,label,x,y,ra,dec = [i.strip() for i in line.replace('\n','').split(',')]

    assert date == dt.datetime.now(dt.timezone.utc).date().isoformat()
    assert name == app.image.name
    assert group == "None"
    assert label == "None"
    assert x == "nan"
    assert y == "nan"
    assert ra == "nan"
    assert dec == "nan"

def test_change_mark_group_save(app:MainWindow, qtbot:QtBot):
    app.mark(group=1, test=True)
    date = 0
    name = 0
    group = 0
    label = 0
    x = 0
    y = 0
    ra = 0
    dec = 0
    line0 = True
    for line in open(os.path.join(config.SAVE_DIR,f'{config.USER}_marks.csv')):
        if line0: line0 = False
        else:
            date,name,group,label,x,y,ra,dec = [i.strip() for i in line.replace('\n','').split(',')]

    assert date == dt.datetime.now(dt.timezone.utc).date().isoformat()
    assert name == app.image.name
    assert group == config.GROUP_NAMES[1]
    assert label == "None"
    assert x == str(float(app.image.width/2))
    assert y == str(float(app.image.height/2))
    assert ra == str(app.image.marks[0].wcs_center[0])
    assert dec == str(app.image.marks[0].wcs_center[1])

    new_group = "BCG"

    config.GROUP_NAMES[1] = new_group
    app.settings_window.group_boxes[0].setText(new_group)
    app.settings_window.update_config()
    app.save()

    line0 = True
    for line in open(os.path.join(config.SAVE_DIR,f'{config.USER}_marks.csv')):
        if line0: line0 = False
        else:
            date,name,group,label,x,y,ra,dec = [i.strip() for i in line.replace('\n','').split(',')]
    
    assert date == dt.datetime.now(dt.timezone.utc).date().isoformat()
    assert name == app.image.name
    assert group == new_group
    assert label == "1"
    assert x == str(float(app.image.width/2))
    assert y == str(float(app.image.height/2))
    assert ra == str(app.image.marks[0].wcs_center[0])
    assert dec == str(app.image.marks[0].wcs_center[1])

def test_next_image(app:MainWindow, qtbot):
    current_image_array = app.image.array
    app.shift(+1)
    new_image_array = app.image.array

    assert not np.all(current_image_array == new_image_array)

# def test_save_category(app, qtbot):
#     app.mark(group=1, test=True)
#     date = 0
#     name = 0
#     group = 0
#     label = 0
#     x = 0
#     y = 0
#     ra = 0
#     dec = 0
#     line0 = True
#     for line in open(os.path.join(config.SAVE_DIR,f'{config.USER}_marks.csv')):
#         if line0: line0 = False
#         else:
#             date,name,group,label,x,y,ra,dec = [i.strip() for i in line.replace('\n','').split(',')]

#     assert date == dt.datetime.now(dt.timezone.utc).date().isoformat()
#     assert name == app.image.name
#     assert group == config.GROUP_NAMES[1]
#     assert label == "None"
#     assert x == str(float(app.image.width/2))
#     assert y == str(float(app.image.height/2))
#     assert ra == str(app.image.marks[0].wcs_center[0]")
#     assert dec == str(app.image.marks[0].wcs_center[1]")


"""
Copyright © 2025, UChicago Argonne, LLC

Full license found at _YOUR_INSTALLATION_DIRECTORY_/imgmarker/LICENSE
"""

import numpy as np
from astropy.wcs import WCS
import numpy as np

class Angle(np.ndarray):
    """Represents a single angle or an array of angles."""

    def __new__(cls,a):
        obj = np.asarray(a,dtype=float).view(cls)
        return obj

    def __add__(self, value):
        return Angle(super().__add__(value))
    
    def __sub__(self, value):
        return Angle(super().__sub__(value))
    
    def __mul__(self, value):
        return Angle(super().__mul__(value))
    
    def __abs__(self):
        return Angle(super().__abs__())
    
    def __truediv__(self, value):
        return Angle(super().__truediv__(value))
    
    @property
    def hms(self):
        sign = np.sign(self)
        m, s = np.divmod(np.abs(self)*240, 60)
        h, m = np.divmod(m, 60)
        return sign*h, sign*m, sign*s

    @property
    def dms(self):
        sign = np.sign(self)
        m, s = np.divmod(np.abs(self)*3600, 60)
        d, m = np.divmod(m, 60)
        return sign*d, sign*m, sign*s
    
class WorldCoord:
    """Object containing a right ascension and declination or arrays thereof."""

    def __init__(self, ra:Angle|float|int, dec:Angle|float|int):
        self.ra = Angle(ra)
        self.dec = Angle(dec)

    def __iter__(self):
        return iter((self.ra,self.dec))
    
    def __getitem__(self,index):
        return (self.ra,self.dec)[index]
    
    def __len__(self):
        return 2
    
    def topix(self, wcs:WCS) -> 'PixCoord':
        """Converts world coordinate into pixel coordinates. The origin is the upper-left corner."""

        _radec = np.dstack((self.ra,self.dec))[0]
        x, _y = wcs.all_world2pix(_radec, 0).T
        y = wcs.pixel_shape[1] - _y
        
        if len(x) == 1:
            return PixCoord(x[0],y[0])
        else:
            return PixCoord(x,y)


class PixCoord:
    """Object containing an x and y coordinate or arrays thereof."""
    
    def __init__(self, x:int|float, y:int|float):
        self.x = np.round(x)
        self.y = np.round(y)

    def __iter__(self):
        return iter((self.x,self.y))
    
    def __getitem__(self,index):
        return (self.x,self.y)[index]
    
    def __len__(self):
        return 2
    
    def toworld(self,wcs:WCS) -> WorldCoord:
        """Converts pixel coordinate into world coordinates. The origin is the upper-left corner."""

        _x, _y = self.x, wcs.pixel_shape[1] - self.y
        _xy = np.dstack((_x,_y))[0]
        ra, dec = wcs.all_pix2world(_xy, 0).T

        if len(ra) == 1:
            return WorldCoord(ra[0],dec[0])
        else:
            return WorldCoord(ra,dec)


"""
Copyright © 2025, UChicago Argonne, LLC

Full license found at _YOUR_INSTALLATION_DIRECTORY_/imgmarker/LICENSE
"""

from imgmarker.gui.pyqt import QApplication
from imgmarker.gui.mark import *
from imgmarker.gui.widget import *

class Screen:
    @staticmethod
    def width():
        return QApplication.primaryScreen().size().width()
    
    @staticmethod
    def height():
        return QApplication.primaryScreen().size().height()
    
    @staticmethod
    def center():
        return QApplication.primaryScreen().geometry().center()


"""
Copyright © 2025, UChicago Argonne, LLC

Full license found at _YOUR_INSTALLATION_DIRECTORY_/imgmarker/LICENSE
"""

"""This contains the classes for the various windows displayed by Image Marker."""

from imgmarker.gui.pyqt import (
    QApplication, QMainWindow, QPushButton,
    QLabel, QScrollArea,
    QVBoxLayout, QWidget, QHBoxLayout, QLineEdit, 
    QCheckBox, QSlider,
    QLineEdit, QFileDialog, QIcon, QFont, QAction, 
    Qt, QPoint, QSpinBox, QMessageBox, QTableWidget, 
    QTableWidgetItem, QHeaderView, QShortcut,
    QDesktopServices, QUrl, QMenu, QColorDialog, 
    QPen, QBrush, QPixmap, QPainter, PYQT_VERSION_STR
)
from imgmarker.gui import Screen, QHLine, PosWidget, RestrictedLineEdit, DefaultDialog
from imgmarker import HEART_SOLID, HEART_CLEAR, __version__, __license__, __docsurl__
from imgmarker import io, image, config
from imgmarker.coordinates import Angle, PixCoord, WorldCoord
import sys
from math import floor, inf, nan
import numpy as np
from numpy import argsort
from functools import partial
import os
from copy import deepcopy
import gc
import shutil
import warnings

def _open_save() -> str:
    dialog = DefaultDialog()
    dialog.setWindowTitle("Open save directory")
    dialog.exec()
    if dialog.closed: sys.exit()

    save_dir = dialog.selectedFiles()[0]
    return save_dir

def _open_ims() -> str:
    dialog = DefaultDialog(config.SAVE_DIR)
    dialog.setWindowTitle("Open image directory")
    dialog.exec()
    if dialog.closed: sys.exit()

    image_dir = dialog.selectedFiles()[0]
    return image_dir

class SettingsWindow(QWidget):
    """Class for the window for settings."""

    def __init__(self,mainwindow:'MainWindow'):
        super().__init__()
        
        layout = QVBoxLayout()
        self.setWindowTitle('Settings')
        self.setLayout(layout)
        self.mainwindow = mainwindow

        # Groups
        self.group_label = QLabel()
        self.group_label.setAlignment(Qt.AlignmentFlag.AlignCenter)
        self.group_label.setText('Groups')

        self.group_boxes = []
        for i in range(1,10):
            lineedit = RestrictedLineEdit([Qt.Key.Key_Comma])
            lineedit.setPlaceholderText(config.GROUP_NAMES[i])
            lineedit.setFixedHeight(30)
            lineedit.setText(config.GROUP_NAMES[i])
            self.group_boxes.append(lineedit)

        self.group_layout = QHBoxLayout()
        for box in self.group_boxes: self.group_layout.addWidget(box)

        # Max marks per group
        self.max_label = QLabel()
        self.max_label.setAlignment(Qt.AlignmentFlag.AlignCenter)
        self.max_label.setText('Max marks per group')

        self.max_boxes = []
        for i in range(0,9):
            spinbox = QSpinBox()
            spinbox.setSpecialValueText('-')
            spinbox.setFixedHeight(30)
            spinbox.setMaximum(9)
            value:str = config.GROUP_MAX[i]
            if value.isnumeric(): spinbox.setValue(int(value))
            spinbox.valueChanged.connect(self.update_config)
            self.max_boxes.append(spinbox)

        self.max_layout = QHBoxLayout()
        for box in self.max_boxes: self.max_layout.addWidget(box)

        # Categories
        self.category_label = QLabel()
        self.category_label.setAlignment(Qt.AlignmentFlag.AlignCenter)
        self.category_label.setText('Categories')

        self.category_boxes = []
        for i in range(1,6):
            lineedit = RestrictedLineEdit([Qt.Key.Key_Comma])
            lineedit.setPlaceholderText(config.CATEGORY_NAMES[i])
            lineedit.setFixedHeight(30)
            lineedit.setText(config.CATEGORY_NAMES[i])
            self.category_boxes.append(lineedit)

        self.category_layout = QHBoxLayout()
        for box in self.category_boxes: self.category_layout.addWidget(box)

        # Options
        self.show_sexagesimal_box = QCheckBox(text='Show sexagesimal coordinates of cursor', parent=self)
        if self.mainwindow.image.wcs == None:
            self.show_sexagesimal_box.setEnabled(False)
        else:
            self.show_sexagesimal_box.setEnabled(True)

        self.focus_box = QCheckBox(text='Middle-click to focus centers the cursor', parent=self)
        
        self.randomize_box = QCheckBox(text='Randomize order of images', parent=self)
        self.randomize_box.setChecked(config.RANDOMIZE_ORDER)

        self.duplicate_box = QCheckBox(text='Insert duplicate images for testing user consistency', parent=self)
        self.duplicate_box.setChecked(False)
        try:
            self.duplicate_box.checkStateChanged.connect(self.duplicate_percentage_state)
        except:
            self.duplicate_box.stateChanged.connect(self.duplicate_percentage_state)

        horizontal_duplicate_layout = QHBoxLayout()

        self.duplicate_percentage_label = QLabel()
        self.duplicate_percentage_label.setAlignment(Qt.AlignmentFlag.AlignLeft)
        self.duplicate_percentage_label.setText("Percentage of dataset to duplicate:")
        
        self.duplicate_percentage_spinbox = QSpinBox()
        self.duplicate_percentage_spinbox.setFixedHeight(25)
        self.duplicate_percentage_spinbox.setFixedWidth(50)
        self.duplicate_percentage_spinbox.setRange(1,100)
        
        self.duplicate_percentage_spinbox.setAlignment(Qt.AlignmentFlag.AlignLeft)
        self.duplicate_percentage_spinbox.valueChanged.connect(self.update_duplicate_percentage)

        if not self.duplicate_box.isChecked():
            self.duplicate_percentage_spinbox.setEnabled(False)
        else:
            self.duplicate_percentage_spinbox.setEnabled(True)
        horizontal_duplicate_layout.setContentsMargins(0,0,345,0)
        horizontal_duplicate_layout.addWidget(self.duplicate_percentage_label)
        horizontal_duplicate_layout.addWidget(self.duplicate_percentage_spinbox)

        # Main layout
        layout.addWidget(self.group_label)
        layout.addLayout(self.group_layout)
        layout.addWidget(self.max_label)
        layout.addLayout(self.max_layout)
        layout.addWidget(QHLine())
        layout.addWidget(self.category_label)
        layout.addLayout(self.category_layout)
        layout.addWidget(QHLine())
        layout.addWidget(self.show_sexagesimal_box)
        layout.addWidget(self.focus_box)
        layout.addWidget(self.randomize_box)
        layout.addWidget(self.duplicate_box)
        layout.addLayout(horizontal_duplicate_layout)
        layout.addWidget(QHLine())
        layout.setAlignment(Qt.AlignmentFlag.AlignCenter)
        self.setFixedWidth(int(Screen.width()/3))
        self.setFixedHeight(layout.sizeHint().height())

        # Set position of window
        qt_rectangle = self.frameGeometry()
        qt_rectangle.moveCenter(Screen.center())
        self.move(qt_rectangle.topLeft())

    def show(self):
        super().show()
        self.activateWindow()

    def keyPressEvent(self, event):
        if event.key() == Qt.Key.Key_Return:
            for box in self.group_boxes: box.clearFocus()
            for box in self.category_boxes: box.clearFocus()
            for box in self.max_boxes: box.clearFocus()

            self.update_config()

        return super().keyPressEvent(event)
    
    def closeEvent(self, a0):
        for box in self.group_boxes: box.clearFocus()
        for box in self.category_boxes: box.clearFocus()
        for box in self.max_boxes: box.clearFocus()

        if self.isVisible():
            fix_over_limit = self.check_max_marks()

            if fix_over_limit:
                a0.ignore()
                return
        else:
            self.update_config()
            self.mainwindow.save()
            self.mainwindow.centralWidget().setFocus()
            return super().closeEvent(a0)
    
    def check_max_marks(self):
        marks_in_group = []
        over_limit_groups = []
        popup_message = ""
        for i, spinbox in enumerate(self.max_boxes):
            limit = spinbox.value()
            if limit > 0:
                for image in self.mainwindow.images:
                    group = i + 1
                    if image.duplicate == True:
                        marks = image.dupe_marks
                    else:
                        marks = image.marks

                    marks_in_group = [mark for mark in marks if mark.g == (group)]

                    if len(marks_in_group) > limit:
                        over_limit_groups.append(group)

        over_limit_groups = np.sort(list(set(over_limit_groups)))
        
        if len(over_limit_groups) == 1:
            popup_message = f"You have set a mark limit for group {[config.GROUP_NAMES[i] for i in over_limit_groups]} that is lower than the number of marks you have placed in that group. Would you like to fix this (click Yes) or continue (click No)?\n\n" \
            "If you continue, the previously placed marks will still be above the limit until you fix the limit or delete the excess marks."
            reply = QMessageBox.question(self, 'WARNING',
                popup_message, QMessageBox.StandardButton.Yes, QMessageBox.StandardButton.No)

            if reply == QMessageBox.StandardButton.Yes:
                return True
            else:
                return False
            
        elif len(over_limit_groups) > 1:
            popup_message = f"You have set a mark limit for groups {[config.GROUP_NAMES[i] for i in over_limit_groups]} that is lower than the number of marks you placed in those groups. Would you like to fix this (click Yes) or continue (click No)?\n\n" \
            "If you continue, the previously placed marks will still be above the limit until you fix the limit or delete the excess marks."
            reply = QMessageBox.question(self, 'WARNING',
                popup_message, QMessageBox.StandardButton.Yes, QMessageBox.StandardButton.No)

            if reply == QMessageBox.StandardButton.Yes:
                return True
            else:
                return False

    def duplicate_percentage_state(self):
        if not self.duplicate_box.isChecked():
            self.duplicate_percentage_spinbox.setEnabled(False)
        else:
            self.duplicate_percentage_spinbox.setEnabled(True)

    def update_duplicate_percentage(self):
        percentage = self.duplicate_percentage_spinbox.value()
        self.mainwindow.update_duplicates(percentage)

    def update_config(self):
        group_names_old = config.GROUP_NAMES.copy()

        # Get the new settings from the boxes
        config.GROUP_NAMES = ['None'] + [box.text() for box in self.group_boxes]
        config.GROUP_MAX = [str(box.value()) if box.value() != 0 else 'None' for box in self.max_boxes]
        config.CATEGORY_NAMES = ['None'] + [box.text() for box in self.category_boxes]
        config.RANDOMIZE_ORDER = self.randomize_box.isChecked()

        for i, box in enumerate(self.mainwindow.category_boxes): 
            box.setText(config.CATEGORY_NAMES[i+1])
            box.setShortcut(self.mainwindow.category_shortcuts[i])
            
        # Update mark labels that haven't been changed
        for image in self.mainwindow.images:
            if image.duplicate == True: marks = image.dupe_marks
            else: marks = image.marks
            for mark in marks:
                try:
                    if mark.label.lineedit.text() in group_names_old:
                        mark.label.lineedit.setText(config.GROUP_NAMES[mark.g])
                except: pass

        # Update text in the controls window 
        self.mainwindow.controls_window.update_text()

        # Save the new settings into the config file
        config.update()

class BlurWindow(QWidget):
    """Class for the blur adjustment window."""

    def __init__(self):
        super().__init__()
        
        layout = QVBoxLayout()
        self.setWindowTitle('Gaussian Blur')
        self.setLayout(layout)

        self.slider = QSlider()
        self.slider.setMinimum(0)
        self.slider.setTickInterval(1)
        self.slider.setSingleStep(1)
        self.slider.setOrientation(Qt.Orientation.Horizontal)
        self.slider.valueChanged.connect(self.slider_moved) 
        self.slider.setPageStep(0)

        self.value_label = QLabel()
        self.value_label.setAlignment(Qt.AlignmentFlag.AlignCenter)
        self.value_label.setText(f'Radius: {self.slider.value()}')

        layout.addWidget(self.value_label)
        layout.addWidget(self.slider)
        layout.setAlignment(Qt.AlignmentFlag.AlignCenter)
        self.setFixedWidth(int(Screen.width()/6))
        self.setFixedHeight(layout.sizeHint().height())

        # Set position of window
        qt_rectangle = self.frameGeometry()
        qt_rectangle.moveCenter(Screen.center())
        self.move(qt_rectangle.topLeft())

    def slider_moved(self, pos):
        self.value_label.setText(f'Radius: {floor(pos)/2}')

    def show(self):
        super().show()
        self.activateWindow()

class FrameWindow(QWidget):
    """Class for the window for switching between frames in an image."""

    def __init__(self):
        super().__init__()
        
        layout = QVBoxLayout()
        self.setWindowTitle('Frames')
        self.setLayout(layout)

        self.slider = QSlider()
        self.slider.setMinimum(0)
        self.slider.setTickInterval(1)
        self.slider.setSingleStep(1)
        self.slider.setOrientation(Qt.Orientation.Horizontal)
        self.slider.sliderMoved.connect(self.slider_moved) 
        self.slider.valueChanged.connect(self.value_changed)

        self.value_label = QLabel()
        self.value_label.setAlignment(Qt.AlignmentFlag.AlignCenter)
        self.value_label.setText(f'Frame: {self.slider.value()}')

        layout.addWidget(self.value_label)
        layout.addWidget(self.slider)
        layout.setAlignment(Qt.AlignmentFlag.AlignCenter)
        self.setFixedWidth(int(Screen.width()/6))
        self.setFixedHeight(layout.sizeHint().height())

        # Set position of window
        qt_rectangle = self.frameGeometry()
        qt_rectangle.moveCenter(Screen.center())
        self.move(qt_rectangle.topLeft())

    def slider_moved(self, pos):
        self.slider.setValue(floor(pos))

    def value_changed(self,value):
        self.value_label.setText(f'Frame: {self.slider.value()}')

    def show(self):
        super().show()
        self.activateWindow()

class ControlsWindow(QWidget):
    """Class for the window that displays the controls."""

    def __init__(self):
        super().__init__()

        layout = QVBoxLayout()
        self.setLayout(layout)
        self.setWindowTitle('Controls')

        self.table = QTableWidget()
        self.table.setColumnCount(2)
        self.table.setSizeAdjustPolicy(QScrollArea.SizeAdjustPolicy.AdjustToContents)
        self.table.verticalHeader().setVisible(False)
        self.table.horizontalHeader().setVisible(False)  

        self.update_text()

        layout.addWidget(self.table)

        # Resize window according to size of layout
        self.resize(int(Screen.width()*0.25), self.sizeHint().height())
        self.setMaximumHeight(self.height())
        
    def update_text(self):
        # Lists for keybindings
        actions_list = ['Next','Back','Change frame','Delete mark','Delete selected marks','Enter comment', 'Focus', 'Zoom in/out', 'Zoom to fit', 'Copy selected mark coordinates' ,'Favorite']
        group_list = [f'Group \"{group}\"' for group in config.GROUP_NAMES[1:]]
        category_list = [f'Category \"{category}\"' for category in config.CATEGORY_NAMES[1:]]
        actions_list = group_list + category_list + actions_list
        buttons_list = ['1 OR Left Click', '2', '3', '4', '5', '6', '7', '8', '9', 'Ctrl+1', 'Ctrl+2', 'Ctrl+3', 'Ctrl+4', 'Ctrl+5', 'Tab', 'Shift+Tab', 'Spacebar', 'Shift+Left Click', 'Delete', 'Enter', 'Middle Click', 'Scroll Wheel', 'Ctrl + 0', 'Ctrl + C', 'F']
        
        items = [ (action, button) for action, button in zip(actions_list, buttons_list) ]

        self.table.setRowCount(len(actions_list))

        for i, (action, button) in enumerate(items):
            action_item = QTableWidgetItem(action)
            button_item = QTableWidgetItem(button)

            action_item.setFlags(Qt.ItemFlag.ItemIsEnabled)
            button_item.setFlags(Qt.ItemFlag.ItemIsEnabled)

            self.table.setItem(i, 0, action_item)
            self.table.setItem(i, 1, button_item)
        
        self.table.horizontalHeader().setSectionResizeMode(QHeaderView.ResizeMode.Stretch)
    
    def show(self):
        """Shows the window and moves it to the front."""

        super().show()
        self.activateWindow()

class AboutWindow(QWidget):
    """Class for the window that displays information about Image Marker."""

    def __init__(self):
        super().__init__()

        layout = QVBoxLayout()
        self.setWindowTitle('About')
        self.setLayout(layout)

        # Create text
        font = QFont('Courier')
        self.layouts = [QHBoxLayout(),QHBoxLayout(),QHBoxLayout(),QHBoxLayout()]
        params = ['Version','PyQt Version','License','Authors']
        labels = [QLabel(f'<div>{__version__}</div>'),
                  QLabel(f'<div>{PYQT_VERSION_STR}</div>'),
                  QLabel(f'<div><a href="https://opensource.org/license/mit">{__license__}</a></div>'),
                  QLabel(f'<div>Andi Kisare, Ryan Walker, and Lindsey Bleem</div>')]

        for label, param in zip(labels, params):
            param_layout = QHBoxLayout()

            param_label = QLabel(f'{param}:')
            param_label.setAlignment(Qt.AlignmentFlag.AlignLeft)
            param_label.setFont(font)

            label.setAlignment(Qt.AlignmentFlag.AlignLeft)
            label.setFont(font)
            if param != 'License': label.setTextInteractionFlags(Qt.TextInteractionFlag.TextSelectableByMouse)
            else: label.setOpenExternalLinks(True)

            param_layout.addWidget(param_label)
            param_layout.addWidget(label)
            param_layout.addStretch(1)
            layout.addLayout(param_layout)

        # Add scroll area to layout, get size of layout
        layout_width, layout_height = layout.sizeHint().width(), layout.sizeHint().height()

        # Resize window according to size of layout
        self.setFixedSize(int(layout_width*1.1),int(layout_height*1.1))       

        # Set position of window
        qt_rectangle = self.frameGeometry()
        qt_rectangle.moveCenter(Screen.center())
        self.move(qt_rectangle.topLeft()) 

    def show(self):
        """Shows the window and moves it to the front."""

        super().show()
        self.activateWindow()

class MarkMenu(QMenu):
    def __init__(self,mainwindow:'MainWindow'):
        super().__init__()
        self.menus:dict[str,QMenu] = {}
        self.mainwindow = mainwindow
        self.setTitle('Mark')

    def menu_setup(self,path:str):
        file = path.split(os.sep)[-1]

        if path == self.mainwindow.markfile.path:
            self.menus[path] = QMenu(f'{file} (default)')
        else:
            self.menus[path] = QMenu(f'{file}')

        # Toggle marks
        marks_action = QAction('Show Marks', self)
        marks_action.setCheckable(True)
        marks_action.setChecked(True)
        marks_action.triggered.connect(partial(self.mainwindow.toggle_marks,path))
        self.menus[path].addAction(marks_action)
        
        ### Toggle mark labels menu
        labels_action = QAction('Show Mark Labels', self)
        labels_action.setCheckable(True)
        labels_action.setChecked(True)
        labels_action.triggered.connect(partial(self.mainwindow.toggle_mark_labels,path))
        self.menus[path].addAction(labels_action)

        if self.mainwindow.n_marks(path) == 0:
            marks_action.setEnabled(False)
            labels_action.setEnabled(False)
        else:
            marks_action.setEnabled(True)
            labels_action.setEnabled(True)

        self.menus[path].addSeparator()

        color_action = QAction('Default Color...', self)
        color_action.triggered.connect(partial(self.mainwindow.update_colors,path))
        color_action.setToolTip('Edit color of marks that aren\'t part of a group')
        self.menus[path].addAction(color_action)
        self.update_color(path)

        if len(self.mainwindow.imageless_marks) == 0:
            color_action.setEnabled(False)
        else:
            color_action.setEnabled(True)

        self.menus[path].addSeparator()

        if path == self.mainwindow.markfile.path:            
            del_marks_action = QAction(f'Delete Marks in Current Image', self)
            
            del_marks_action.triggered.connect(partial(self.mainwindow.del_usermarks,'all'))
            self.menus[path].addAction(del_marks_action)

        else:
            del_file_action = QAction(f'Delete', self)
            del_file_action.triggered.connect(partial(self.mainwindow.del_markfile,path))
            self.menus[path].addAction(del_file_action)

        self.addMenu(self.menus[path])

    def update_menu(self,path:str):
        if self.mainwindow.n_marks(path) == 0:
            self.marks_action(path).setEnabled(False)
            self.labels_action(path).setEnabled(False)
        else:
            self.marks_action(path).setEnabled(True)
            self.labels_action(path).setEnabled(True)

        if len(self.mainwindow.imageless_marks) == 0:
            self.color_action(path).setEnabled(False)
        else:
            self.color_action(path).setEnabled(True)

    def update_color(self,path):
        s = 14
        pixmap = QPixmap(s,s)
        painter = QPainter(pixmap)
        
        pen = QPen(Qt.GlobalColor.black, 2)
        painter.setBrush(config.DEFAULT_COLORS[path])
        painter.setPen(pen)
        painter.drawRect(0, 0, s, s)
        painter.end()
        icon = QIcon(pixmap)
        self.color_action(path).setIcon(icon)

    def color_action(self,path):
        return [action for action in self.menus[path].actions() if action.text() == "Default Color..."][0]
    
    def marks_action(self,path):
        return [action for action in self.menus[path].actions() if action.text() == "Show Marks"][0]
    
    def labels_action(self,path):
        return [action for action in self.menus[path].actions() if action.text() == "Show Mark Labels"][0]
        
class MainWindow(QMainWindow):
    """Class for the main window."""

    def __init__(self):
        super().__init__()

        self.setWindowTitle("Image Marker")
        self.frame = 0
        self.markfile = io.MarkFile(os.path.join(config.SAVE_DIR,f'{config.USER}_marks.csv'))
        self.imagesfile = io.ImagesFile()
        self.favoritesfile = io.FavoritesFile()
        
        # Shortcuts
        del_shortcuts = [QShortcut('Backspace', self), QShortcut('Delete', self)]
        for shortcut in del_shortcuts: shortcut.activated.connect(self.del_usermarks)

        shiftplus_shorcut = QShortcut('Space', self)
        shiftplus_shorcut.activated.connect(partial(self.shiftframe,1))

        shiftminus_shorcut = QShortcut('Shift+Space', self)
        shiftminus_shorcut.activated.connect(partial(self.shiftframe,-1))

        ctrlc_shortcut = QShortcut('Ctrl+C', self)
        ctrlc_shortcut.activated.connect(self.copy_to_clipboard)
    
        # Initialize data
        self.order = []
        self.__init_data__()
        self.image_scene = image.ImageScene(self.image)
        self.image_view = image.ImageView(self.image_scene)
        self.image_view.mouseMoveEvent = self.mouseMoveEvent
        self.clipboard = QApplication.clipboard()

        #Initialize inserting duplicates at random
        self.images_seen_since_duplicate_count = 0 #keeps track of how many images have been seen since last duplicate
        self.duplicate_image_interval = 1 #this will vary every time a duplicate image is seen
        self.duplicates_seen = []
        self.rng = np.random.default_rng()

        # Setup child windows
        self.blur_window = BlurWindow()
        self.blur_window.slider.sliderReleased.connect(partial(self.image.blur,self.blur_window.slider.sliderPosition))
        
        self.frame_window = FrameWindow()
        self.frame_window.slider.valueChanged.connect(self.image.seek)
        self.frame_window.slider.setMaximum(self.image.n_frames-1)

        self.settings_window = SettingsWindow(self)
        # self.settings_window.show_sexagesimal_box.stateChanged.connect(self.show_sexagesimal)
        self.settings_window.focus_box.stateChanged.connect(partial(setattr,self.image_view,'cursor_focus'))
        self.settings_window.randomize_box.stateChanged.connect(self.toggle_randomize)

        self.controls_window = ControlsWindow()
        self.about_window = AboutWindow()

        # Update max blur
        self.blur_window.slider.setMaximum(self.blur_max)

        # Current image widget
        self.image_label = QLabel(f'{self.image.name} ({self.idx+1} of {self.N})')
        self.image_label.setTextInteractionFlags(Qt.TextInteractionFlag.TextSelectableByMouse)
        self.image_label.setAlignment(Qt.AlignmentFlag.AlignHCenter)

        # Mouse position widget
        self.pos_widget = PosWidget()
        if self.image.wcs == None: 
            self.pos_widget.hidewcs()
        else:
            self.pos_widget.showwcs()

        # Back widget
        self.back_button = QPushButton(text='Back',parent=self)
        self.back_button.setFixedHeight(40)
        self.back_button.clicked.connect(partial(self.shift,-1))
        self.back_button.setShortcut('Shift+Tab')
        self.back_button.setFocusPolicy(Qt.FocusPolicy.NoFocus)

        # Enter Button
        self.submit_button = QPushButton(text='Enter',parent=self)
        self.submit_button.setFixedHeight(40)
        self.submit_button.clicked.connect(self.enter)
        # self.submit_button.setShortcut('Return')
        self.submit_button.setFocusPolicy(Qt.FocusPolicy.NoFocus)

        # Next widget
        self.next_button = QPushButton(text='Next',parent=self)
        self.next_button.setFixedHeight(40)
        self.next_button.clicked.connect(partial(self.shift,1))
        self.next_button.setShortcut('Tab')
        self.next_button.setFocusPolicy(Qt.FocusPolicy.NoFocus)

        # Comment widget
        self.comment_box = QLineEdit(parent=self)
        self.comment_box.setFixedHeight(40)
    
        # Botton Bar layout
        self.bottom_layout = QHBoxLayout()
        self.bottom_layout.addWidget(self.back_button)
        self.bottom_layout.addWidget(self.next_button)
        self.bottom_layout.addWidget(self.comment_box)
        self.bottom_layout.addWidget(self.submit_button)
        
        ### Category widgets
        self.categories_layout = QHBoxLayout()

        # Category boxes
        self.category_shortcuts = ['Ctrl+1', 'Ctrl+2', 'Ctrl+3', 'Ctrl+4', 'Ctrl+5']
        self.category_boxes = [QCheckBox(text=config.CATEGORY_NAMES[i], parent=self) for i in range(1,6)]
        for i, box in enumerate(self.category_boxes):
            box.setFixedHeight(20)
            box.setStyleSheet("margin-left:30%; margin-right:30%;")
            box.clicked.connect(partial(self.categorize,i+1))
            box.setFocusPolicy(Qt.FocusPolicy.NoFocus)
            box.setShortcut(self.category_shortcuts[i])
            self.categories_layout.addWidget(box)

        # Favorite box
        # self.favorite_list = self.favoritesfile.read()
        self.favorite_box = QCheckBox(parent=self)
        self.favorite_box.setFixedHeight(20)
        self.favorite_box.setFixedWidth(40)
        self.favorite_box.setIcon(QIcon(HEART_CLEAR))
        self.favorite_box.setTristate(False)
        self.favorite_box.clicked.connect(self.favorite)
        self.favorite_box.setFocusPolicy(Qt.FocusPolicy.NoFocus)
        self.categories_layout.addWidget(self.favorite_box)
        self.favorite_box.setShortcut('F')

        # Add widgets to main layout
        central_widget = QWidget()
        layout = QVBoxLayout(central_widget)
        layout.addWidget(self.image_label)
        layout.addWidget(self.image_view)
        layout.addWidget(self.pos_widget)
        layout.addWidget(QHLine())
        layout.addLayout(self.bottom_layout)
        layout.addLayout(self.categories_layout)
        self.setCentralWidget(central_widget)
        
        # Menu bar
        menubar = self.menuBar()

        ## File menu
        file_menu = menubar.addMenu("File")

        ### Open file menu
        open_action = QAction('Open Save...', self)
        open_action.setShortcuts(['Ctrl+o'])
        open_action.triggered.connect(self.open)
        file_menu.addAction(open_action)

        ### Open new image folder menu
        import_ims_action = QAction('Open Images...', self)
        import_ims_action.setShortcuts(['Ctrl+Shift+i'])
        import_ims_action.triggered.connect(self.import_ims)
        file_menu.addAction(import_ims_action)

        file_menu.addSeparator()

        ### Import mark file
        import_marks_action = QAction('Import Mark File...', self)
        import_marks_action.setShortcuts(['Ctrl+Shift+m'])
        import_marks_action.triggered.connect(self.import_markfile)
        file_menu.addAction(import_marks_action)
        
        ### Exit menu
        file_menu.addSeparator()
        exit_action = QAction('Exit', self)
        exit_action.setShortcuts(['Ctrl+q'])
        exit_action.triggered.connect(self.close)
        file_menu.addAction(exit_action)

        ## Edit menu
        edit_menu = menubar.addMenu("Edit")
        edit_menu.setToolTipsVisible(True)

        ### Undo previous mark
        undo_mark_action = QAction('Undo Previous Mark', self)
        undo_mark_action.setShortcuts(['Ctrl+z'])
        undo_mark_action.triggered.connect(self.undo_prev_mark)
        edit_menu.addAction(undo_mark_action)

        ### Redo previous mark
        redo_mark_action = QAction('Redo Previous Mark', self)
        redo_mark_action.setShortcuts(['Ctrl+Shift+z'])
        redo_mark_action.triggered.connect(self.redo_prev_mark)
        edit_menu.addAction(redo_mark_action)

        ### Settings menu
        edit_menu.addSeparator()
        settings_action = QAction('Settings...', self)
        settings_action.setShortcuts(['Ctrl+,'])
        settings_action.triggered.connect(self.settings_window.show)
        edit_menu.addAction(settings_action)

        ## View menu
        view_menu = menubar.addMenu("View")

        ### Zoom menu
        zoom_menu = view_menu.addMenu("Zoom")

        #### Zoom in
        zoomin_action = QAction('Zoom In', self)
        zoomin_action.setShortcuts(['Ctrl+='])
        zoomin_action.triggered.connect(partial(self.image_view.zoom,1.2,'viewport'))
        zoom_menu.addAction(zoomin_action)

        ### Zoom out
        zoomout_action = QAction('Zoom Out', self)
        zoomout_action.setShortcuts(['Ctrl+-'])
        zoomout_action.triggered.connect(partial(self.image_view.zoom,1/1.2,'viewport'))
        zoom_menu.addAction(zoomout_action)

        ### Zoom to Fit
        zoomfit_action = QAction('Zoom to Fit', self)
        zoomfit_action.setShortcuts(['Ctrl+0'])
        zoomfit_action.triggered.connect(self.image_view.zoomfit)
        zoom_menu.addAction(zoomfit_action)

        ### Frame menu
        view_menu.addSeparator()
        self.frame_action = QAction('Frames...', self)
        self.frame_action.setShortcuts(['Ctrl+f'])
        self.frame_action.triggered.connect(self.frame_window.show)
        view_menu.addAction(self.frame_action)

        if self.image.n_frames > 1:
            self.frame_action.setEnabled(True)
        else:
            self.frame_action.setEnabled(False)

        view_menu.addSeparator()

        ## Filter menu
        filter_menu = menubar.addMenu("Filter")

        ### Blur
        blur_action = QAction('Gaussian Blur...',self)
        blur_action.setShortcuts(['Ctrl+b'])
        blur_action.triggered.connect(self.blur_window.show)
        filter_menu.addAction(blur_action)

        ### Scale menus
        filter_menu.addSeparator()
        stretch_menu = filter_menu.addMenu('Stretch')

        linear_action = QAction('Linear', self)
        linear_action.setCheckable(True)
        linear_action.setChecked(True)
        stretch_menu.addAction(linear_action)

        log_action = QAction('Log', self)
        log_action.setCheckable(True)
        stretch_menu.addAction(log_action)

        linear_action.triggered.connect(partial(setattr,self,'stretch',image.Stretch.LINEAR))
        linear_action.triggered.connect(partial(linear_action.setChecked,True))
        linear_action.triggered.connect(partial(log_action.setChecked,False))

        log_action.triggered.connect(partial(setattr,self,'stretch',image.Stretch.LOG))
        log_action.triggered.connect(partial(linear_action.setChecked,False))
        log_action.triggered.connect(partial(log_action.setChecked,True))

        ### Interval menus
        interval_menu = filter_menu.addMenu('Interval')

        minmax_action = QAction('Min-Max', self)
        minmax_action.setCheckable(True)
        minmax_action.setChecked(True)
        interval_menu.addAction(minmax_action)

        zscale_action = QAction('ZScale', self)
        zscale_action.setCheckable(True)
        interval_menu.addAction(zscale_action)

        minmax_action.triggered.connect(partial(setattr,self,'interval',image.Interval.MINMAX))
        minmax_action.triggered.connect(partial(minmax_action.setChecked,True))
        minmax_action.triggered.connect(partial(zscale_action.setChecked,False))

        zscale_action.triggered.connect(partial(setattr,self,'interval',image.Interval.ZSCALE))
        zscale_action.triggered.connect(partial(minmax_action.setChecked,False))
        zscale_action.triggered.connect(partial(zscale_action.setChecked,True))

        ### Marks Menu
        self.mark_menu = MarkMenu(self)
        self.mark_menu.setToolTipsVisible(True)
        for path in io.markpaths():
            self.mark_menu.menu_setup(path)

        menubar.addMenu(self.mark_menu)

        ## Help menu
        help_menu = menubar.addMenu('Help')

        ### Controls window
        controls_action = QAction('Controls', self)
        controls_action.setShortcuts(['F1'])
        controls_action.triggered.connect(self.controls_window.show)
        help_menu.addAction(controls_action)

        ### Documentation
        docs_action = QAction('Documentation', self)
        docs_action.triggered.connect(partial(QDesktopServices.openUrl,QUrl(__docsurl__)))
        help_menu.addAction(docs_action)

        ### About window
        help_menu.addSeparator()
        about_action = QAction('About', self)
        about_action.triggered.connect(self.about_window.show)
        help_menu.addAction(about_action)
        
        # Resize and center MainWindow; move controls off to the right
        self.resize(int(Screen.height()*0.8),int(Screen.height()*0.8))
        
        center = Screen.center()
        center -= QPoint(self.width(),self.height())/2
        self.move(center)

        self.controls_window.move(int(self.x()+self.width()*1.04),self.y())

        # Initialize some data
        self.get_comment()
        self.update_marks()
        self.update_categories()
        self.settings_window.update_duplicate_percentage()

    def __init_data__(self):
        """Initializes images."""
        
        self.images, self.imageless_marks = self.markfile.read(self.imagesfile.read())
        self.favorite_list = self.favoritesfile.read()

        try: self.image.close()
        except: pass
        
        # Find all images in image directory
        try:
            self.images, self.idx = io.glob(edited_images=self.images)
            self.image = self.images[self.idx]
            self.image.seek(self.frame)
            self.image.seen = True
            self.N = len(self.images)
            if self.image.name not in self.order:
                self.order.append(self.image.name)
        except:
            config.IMAGE_DIR = _open_ims()
            if config.IMAGE_DIR == None: sys.exit()
            config.update()
            
            self.images, self.idx = io.glob(edited_images=self.images)
            self.image = self.images[self.idx]
            self.image.seek(self.frame)
            self.image.seen = True
            self.N = len(self.images)
            if self.image.name not in self.order:
                self.order.append(self.image.name)

        # Add marks from imports
        for path in io.markpaths():
            if path != self.markfile.path:
                try:
                    self.images, imageless_marks = io.MarkFile(path).read(self.images)
                    self.imageless_marks += imageless_marks
                except Exception as e:
                    print(f"WARNING: {str(e).strip("'")} Skipping import.")
                    os.remove(path)
    
    @property
    def interval(self): return self._interval_str
    @interval.setter
    def interval(self,value):
        self._interval_str = value
        for img in self.images: img.interval = value
        self.image.rescale()
        
    @property
    def stretch(self): return self._stretch_str
    @stretch.setter
    def stretch(self,value):
        self._stretch_str = value
        for img in self.images: img.stretch = value
        self.image.rescale()

    @property
    def blur_max(self):
        _blur_max = int((self.image.height+self.image.width)/20)
        _blur_max = 10*round(_blur_max/10)
        return max(10, _blur_max)
    
    def n_marks(self,path):
        marks = [mark for mark in self.image.marks if mark.dst == path]
        marks += [mark for mark in self.imageless_marks if mark.dst == path]
        return len(marks)

    def inview(self,x:int|float,y:int|float):
        """
        Checks if x and y are contained within the image.

        Parameters
        ----------
        x: int OR float
            x coordinate
        y: int OR float
            y coordinate

        Returns
        ----------
        True if the (x,y) is contained within the image, False otherwise.
        """

        return (0 <= x) & (x <=self.image.width-1) & (0 <= y) & ( y <= self.image.height-1)

    # === Events ===

    def keyPressEvent(self, a0):
        """Checks which keyboard button was pressed and calls the appropriate function."""

        # Keybinds for show/hide mark file
        modifiers = QApplication.keyboardModifiers()
        alt = modifiers == Qt.KeyboardModifier.AltModifier
        nomod = modifiers == Qt.KeyboardModifier.NoModifier
        
        if alt:
            toggle_mark_keys = [
                Qt.Key.Key_M,Qt.Key.Key_1,Qt.Key.Key_2,
                Qt.Key.Key_3,Qt.Key.Key_4,Qt.Key.Key_5,
                Qt.Key.Key_6,Qt.Key.Key_7,Qt.Key.Key_8,
                Qt.Key.Key_9
            ]

            paths = list(self.mark_menu.menus.keys())
            
            for i in range(len(paths)):
                if a0.key() == toggle_mark_keys[i]:
                    marks_action = self.mark_menu.marks_action(paths[i])
                    marks_action.setChecked(not marks_action.isChecked())
                    self.toggle_marks(paths[i])

            if a0.key() == Qt.Key.Key_L:
                labels_action = self.mark_menu.labels_action(paths[0])
                labels_action.setChecked(not labels_action.isChecked())
                self.toggle_mark_labels(paths[0])
        
        elif nomod:
            # Check if key is bound with marking the image
            for group, binds in config.MARK_KEYBINDS.items():
                if a0.key() in binds: self.mark(group=group)

    def mousePressEvent(self, a0):
        """Checks which mouse button was pressed and calls the appropriate function."""

        #super().mousePressEvent(a0)

        modifiers = QApplication.keyboardModifiers()
        leftbutton = a0.button() == Qt.MouseButton.LeftButton
        middlebutton = a0.button() == Qt.MouseButton.MiddleButton
        alt = modifiers == Qt.KeyboardModifier.AltModifier
        shift = modifiers == Qt.KeyboardModifier.ShiftModifier
        nomod = modifiers == Qt.KeyboardModifier.NoModifier

        # Check if key is bound with marking the image
        for group, binds in config.MARK_KEYBINDS.items():
            if (a0.button() in binds) and nomod: self.mark(group=group)

        if middlebutton or (alt and leftbutton): self.image_view.center_cursor()
        if shift and leftbutton: self.del_usermarks(mode='cursor')
        
    def mouseMoveEvent(self, a0):
        """Operations executed when the mouse cursor is moved."""

        super().mouseMoveEvent(a0)
        self.update_pos()

    def closeEvent(self, a0):
        self.update_comments()
        self.save()
        self.about_window.close()
        self.blur_window.close()
        self.frame_window.close()
        self.controls_window.close()
        self.settings_window.close()
        return super().closeEvent(a0)

    # === Actions ===
    def save(self) -> None:
        """Method for saving image data"""
        
        self.markfile.save(self.images,self.imageless_marks)
        self.imagesfile.save(self.images)
        self.favoritesfile.save(self.favorite_list, self.images)

    def open(self) -> None:
        """Method for the open save directory dialog."""

        open_msg = 'This will save all current data in the current save directory and begin saving new data in the newly selected save directory.\
            Customized configuration file data will be kept if there is no available configuration file in the new save directory.\n\nAre you sure you want to continue?'
        reply = QMessageBox.question(self, 'WARNING',
                        open_msg, QMessageBox.StandardButton.Yes, QMessageBox.StandardButton.No)

        if reply == QMessageBox.StandardButton.No: return

        save_dir = QFileDialog.getExistingDirectory(self, 'Open save directory', config.SAVE_DIR)
        if save_dir == '': return

        before_image_dir = config.IMAGE_DIR
        group_names_old = config.GROUP_NAMES.copy()

        config.SAVE_DIR = save_dir
        config.IMAGE_DIR, config.GROUP_NAMES, config.CATEGORY_NAMES, config.GROUP_MAX, config.RANDOMIZE_ORDER = config.read()
        config.update()

        self.markfile = io.MarkFile(os.path.join(config.SAVE_DIR,f'{config.USER}_marks.csv'))
        self.imagesfile = io.ImagesFile()
        self.favoritesfile = io.FavoritesFile()

        after_image_dir = config.IMAGE_DIR

        if before_image_dir != after_image_dir: # if the image directory is different in the new config file, then we need to purge these lists
            del self.order; del self.duplicates_seen; del self.images
            gc.collect()
            self.order = []
            self.images_seen_since_duplicate_count = 0
            self.duplicates_seen = []

        self.images, self.idx = io.glob(edited_images=[])
        self.N = len(self.images)

        for i, box in enumerate(self.category_boxes): 
            box.setText(config.CATEGORY_NAMES[i+1])
            box.setShortcut(self.category_shortcuts[i])
            
        # Update mark labels that haven't been changed
        for image in self.images:
            if image.duplicate == True: marks = image.dupe_marks
            else: marks = image.marks
            for mark in marks:
                try:
                    if mark.label.lineedit.text() in group_names_old:
                        mark.label.lineedit.setText(config.GROUP_NAMES[mark.g])
                except: pass

        self.__init_data__()
        self.settings_window.__init__(self)
        self.update_images()
        self.image_view.zoomfit()
        self.update_marks()
        self.get_comment()
        self.update_categories()
        self.update_comments()
        self.update_favorites()
        self.controls_window.update_text()

    def import_ims(self) -> None:
        """Method for the open image directory dialog."""

        open_msg = 'This will overwrite all data associated with your current images, including all marks.\n\nAre you sure you want to continue?'
        reply = QMessageBox.question(self, 'WARNING', 
                        open_msg, QMessageBox.StandardButton.Yes, QMessageBox.StandardButton.No)

        if reply == QMessageBox.StandardButton.No: return

        image_dir = QFileDialog.getExistingDirectory(self, 'Open image directory', config.SAVE_DIR)
        if image_dir == '': return

        _image_dir = config.IMAGE_DIR
        config.IMAGE_DIR = image_dir

        del self.order; del self.duplicates_seen; del self.images
        gc.collect()
        self.order = []
        self.images_seen_since_duplicate_count = 0
        self.duplicates_seen = []
        
        self.images, self.idx = io.glob(edited_images=[])
        self.N = len(self.images)

        if self.N == 0:
            config.IMAGE_DIR = _image_dir
            return

        config.update()
        self.update_images()
        self.update_marks()
        self.get_comment()
        self.update_categories()
        self.update_comments()

    def import_markfile(self, **kwargs):
        """Method for opening a catalog file."""
        if 'src' not in kwargs:
            src = QFileDialog.getOpenFileName(self, 'Import mark file', config.SAVE_DIR, 'Text files (*.txt *.csv)')[0]
            if src == '': return None
        else:
            src = kwargs['src']
        
        dst = os.path.join(config.SAVE_DIR,'imports')
        mark_dst = shutil.copy(src,dst)
        
        try:
            self.images, imageless_marks = io.MarkFile(mark_dst).read(self.images)
            self.imageless_marks += imageless_marks

            self.update_marks()

        except Exception as e:
            print(f"WARNING: {str(e).strip("'")} Skipping import.")
            os.remove(mark_dst)
            
    def favorite(self,state) -> None:
        """Favorite the current image."""

        state = Qt.CheckState(state)
        if state == Qt.CheckState.PartiallyChecked:
            self.favorite_box.setIcon(QIcon(HEART_SOLID))
            self.favorite_list.append(self.image.name)
            self.favoritesfile.save(self.favorite_list,self.images)
        else:
            self.favorite_box.setIcon(QIcon(HEART_CLEAR))
            if self.image.name in self.favorite_list: 
                self.favorite_list.remove(self.image.name)
            self.favoritesfile.save(self.favorite_list,self.images)

    def categorize(self,i:int) -> None:
        """Categorize the current image."""

        if (self.category_boxes[i-1].checkState() == Qt.CheckState.Checked) and (i not in self.image.categories):
            self.image.categories.append(i)
        elif (i in self.image.categories):
            self.image.categories.remove(i)
        self.save()

    def copy_to_clipboard(self):

        has_wcs = self.image.wcs != None

        if self.image.duplicate == True:
            marks = self.image.dupe_marks
        else:
            marks = self.image.marks

        selected_marks = [mark for mark in marks if mark.isSelected()]

        if len(selected_marks) == 0:
            return 
        else:
            mark_to_copy = selected_marks[-1]

        if has_wcs:
            ra, dec = mark_to_copy.wcs_center
        
            if self.settings_window.show_sexagesimal_box.isChecked():
                ra_h,ra_m,ra_s = ra.hms
                dec_d,dec_m,dec_s = dec.dms

                ra_str = rf'{np.abs(ra_h):02.0f}h {np.abs(ra_m):02.0f}m {np.abs(ra_s):05.2f}s'
                dec_str = f'{np.abs(dec_d):02.0f}° {np.abs(dec_m):02.0f}\' {np.abs(dec_s):05.2f}\"'.replace('-', '')

            else:
                ra_str = f'{ra:03.6f}'
                dec_str = f'{np.abs(dec):02.6f}'

            if dec > 0: dec_str = '+' + dec_str
            else: dec_str = '-' + dec_str
            
            string_copy = ra_str + ", " + dec_str

        else:
            x, y = str(mark_to_copy.center.x), str(mark_to_copy.center.y)
            string_copy = x + ", " + y

        self.clipboard.setText(string_copy)

    def mark(self, group:int=0, test=False) -> None:
        """Add a mark to the current image."""

        if self.image.duplicate == True:
            marks = self.image.dupe_marks
        else:
            marks = self.image.marks

        # get event position and position on image
        if not test:
            pix_pos = self.image_view.mouse_pix_pos()
            x, y = pix_pos.x(), pix_pos.y()
        else: 
            x = self.image.width/2
            y = self.image.height/2
            
        # Mark if hovering over image
        if config.GROUP_MAX[group - 1] == 'None': limit = inf
        else: limit = int(config.GROUP_MAX[group - 1])

        marks_in_group = [m for m in marks if m.g == group]

        try: 
            if len(marks) >= 1: marks[-1].label.enter()
        except: pass

        marks_action = [action for action in self.mark_menu.menus[self.markfile.path].actions() if action.text() == "Show Marks"][0]
        labels_action = [action for action in self.mark_menu.menus[self.markfile.path].actions() if action.text() == "Show Mark Labels"][0]

        if self.inview(x,y) and ((len(marks_in_group) < limit) or limit == 1):            
            mark = self.image_scene.mark(x,y,group=group)
            
            if (limit == 1) and (len(marks_in_group) == 1):
                prev_mark = marks_in_group[0]
                self.image_scene.rmmark(prev_mark)
                marks.remove(prev_mark)
                marks.append(mark)
            
            elif (len(marks_in_group) > limit) and limit == 1:
                self.image_scene.rmmark(mark)

            else: marks.append(mark)

            marks_enabled = marks_action.isChecked()
            labels_enabled = labels_action.isChecked()

            if labels_enabled: mark.label.show()
            else: mark.label.hide()

            if marks_enabled: 
                mark.show()
                if labels_enabled: mark.label.show()
            else: 
                mark.hide()
                mark.label.hide()

            self.save()
        
        
        if len(marks) == 0:
            marks_action.setEnabled(False)
            labels_action.setEnabled(False)
        else:
            marks_action.setEnabled(True)
            labels_action.setEnabled(True)

    def shift(self,delta:int):
        """Move back or forward *delta* number of images."""
        
        # Increment the index
        self.idx += delta
        if self.idx > self.N-1:
            self.idx = 0
        elif self.idx < 0:
            self.idx = self.N-1
        
        self.update_comments()
        self.update_images()
        self.update_marks()
        self.get_comment()
        self.update_categories()
        self.update_favorites()

        self.save()

    def shiftframe(self,delta:int):
        self.image.seek(self.frame+delta)

        self.frame = self.image.frame
        self.frame_window.slider.setValue(self.frame)
            
    def enter(self):
        """Enter the text in the comment box into the image."""

        self.update_comments()
        self.comment_box.clearFocus()
        self.save()

    def undo_prev_mark(self):
        if self.image.duplicate == True:
            marks = self.image.dupe_marks
        else:
            marks = self.image.marks

        marks_action = [action for action in self.mark_menu.menus[self.markfile.path].actions() if action.text() == "Show Marks"][0]
        labels_action = [action for action in self.mark_menu.menus[self.markfile.path].actions() if action.text() == "Show Mark Labels"][0]

        if len(marks) > 0:
            mark = marks[-1]
            self.image.undone_marks.append(mark)
            self.image_scene.rmmark(mark)
            marks.remove(mark)

            marks_enabled = marks_action.isChecked()
            labels_enabled = labels_action.isChecked()

            if labels_enabled: mark.label.show()
            else: mark.label.hide()

            if marks_enabled: 
                mark.show()
                if labels_enabled: mark.label.show()
            else: 
                mark.hide()
                mark.label.hide()
                
        if len(marks) == 0:
            marks_action.setEnabled(False)
            labels_action.setEnabled(False)
        else:
            marks_action.setEnabled(True)
            labels_action.setEnabled(True)

        self.save()

    def redo_prev_mark(self):
        if self.image.duplicate == True:
            marks = self.image.dupe_marks
        else:
            marks = self.image.marks

        try:
            group = self.image.undone_marks[-1].g
        except:
            return

        marks_in_group = [m for m in marks if m.g == group]

        if config.GROUP_MAX[group - 1] == 'None': limit = inf
        else: limit = int(config.GROUP_MAX[group - 1])

        marks_action = [action for action in self.mark_menu.menus[self.markfile.path].actions() if action.text() == "Show Marks"][0]
        labels_action = [action for action in self.mark_menu.menus[self.markfile.path].actions() if action.text() == "Show Mark Labels"][0]

        if (len(self.image.undone_marks) > 0) and ((len(marks_in_group) < limit) or limit == 1):
            mark = self.image.undone_marks[-1]
            self.image_scene.mark(mark)

            if (limit == 1) and (len(marks_in_group) == 1):
                prev_mark = marks_in_group[0]
                self.image_scene.rmmark(prev_mark)
                marks.remove(prev_mark)
                marks.append(mark)
                self.image.undone_marks.remove(mark)

            else:
                marks.append(mark)
                self.image.undone_marks.remove(mark)

            marks_enabled = marks_action.isChecked()
            labels_enabled = labels_action.isChecked()

            if labels_enabled: mark.label.show()
            else: mark.label.hide()

            if marks_enabled: 
                mark.show()
                if labels_enabled: mark.label.show()
            else: 
                mark.hide()
                mark.label.hide()

        if len(marks) == 0:
            marks_action.setEnabled(False)
            labels_action.setEnabled(False)
        else:
            marks_action.setEnabled(True)
            labels_action.setEnabled(True)

        self.save()

    # === Update methods ===
    def update_pos(self):
        # Mark if hovering over image
        pix_pos = self.image_view.mouse_pix_pos()
        x, y = pix_pos.x(), pix_pos.y()

        if self.inview(x,y):
            _x, _y = x, self.image.height - y

            try: ra, dec = self.image.wcs.all_pix2world([[_x, _y]], 0)[0]
            except: ra, dec = nan, nan

            if self.settings_window.show_sexagesimal_box.isChecked():
                ra_h,ra_m,ra_s = Angle(ra).hms
                dec_d,dec_m,dec_s = Angle(dec).dms

                ra_str = rf'{np.abs(ra_h):02.0f}h {np.abs(ra_m):02.0f}m {np.abs(ra_s):05.2f}s'
                dec_str = f'{np.abs(dec_d):02.0f}° {np.abs(dec_m):02.0f}\' {np.abs(dec_s):05.2f}\"'.replace('-', '')

            else:
                ra_str = f'{ra:03.5f}°'
                dec_str = f'{np.abs(dec):02.5f}°'

            if dec > 0: dec_str = '+' + dec_str
            else: dec_str = '-' + dec_str

            self.pos_widget.x_text.setText(f'{x} px')
            self.pos_widget.y_text.setText(f'{y} px')

            self.pos_widget.ra_text.setText(ra_str)
            self.pos_widget.dec_text.setText(dec_str)

        else:
            self.pos_widget.cleartext()

    def update_duplicates(self, percentage):
        self.min_images_til_duplicate = int((len(self.images) - len(self.duplicates_seen)) / (percentage * 4))
        self.max_images_til_duplicate = int((len(self.images) - len(self.duplicates_seen)) / percentage)

    def update_favorites(self):
        """Update favorite boxes based on the contents of favorite_list."""

        if self.image.name in self.favorite_list:
            self.favorite_box.setChecked(True)
            self.favorite_box.setIcon(QIcon(HEART_SOLID))
        else:
            self.favorite_box.setIcon(QIcon(HEART_CLEAR))
            self.favorite_box.setChecked(False)

    def update_images(self):
        """Updates previous image with a new image."""

        # Disconnect sliders from previous image
        try:
            self.blur_window.slider.sliderReleased.disconnect()
            self.frame_window.slider.valueChanged.disconnect(self.image.seek)
        except: pass

        # Update scene
        _w, _h = self.image.width, self.image.height
        try: self.image.close()
        except: pass

        # Randomizing duplicate images to show for consistency of user marks
        if self.settings_window.duplicate_box.isChecked():
            seen_images = [image for image in self.images if (len(image.marks) != 0) and (image.name not in self.duplicates_seen)]
            if self.settings_window.duplicate_box.isChecked():
                if (len(seen_images) > self.min_images_til_duplicate):
                    self.images_seen_since_duplicate_count += 1
                    if (self.images_seen_since_duplicate_count == self.duplicate_image_interval):
                        self.duplicate_image_interval = self.rng.integers(self.min_images_til_duplicate,self.max_images_til_duplicate)
                        self.images_seen_since_duplicate_count = 0
                        duplicate_image_to_show = deepcopy(self.rng.choice(seen_images[0:-1]))
                        duplicate_image_to_show.duplicate = True
                        duplicate_image_to_show.marks.clear()
                        self.images.insert(self.idx,duplicate_image_to_show)
                        self.N = len(self.images)
                        self.duplicates_seen.append(duplicate_image_to_show.name)
        
        # Continue update_images
        self.frame = self.image.frame
        self.image = self.images[self.idx]
        self.image.seek(self.frame)
        self.image_scene.update_image(self.image)
        if self.image.name not in self.order:   # or self.image.duplicate == True: This could be added to preserve order when duplicates are being inserted, but the use case for someone randomizing
                self.order.append(self.image.name)   # who wants to keep the order if duplicates have been seen and then they turn off and back on randomization is quite low

        # Fit back to view if the image dimensions have changed
        if (self.image.width != _w) or (self.image.height != _h): self.image_view.zoomfit()

        # Update position widget
        self.update_pos()
        if self.image.wcs == None: 
            self.pos_widget.hidewcs()
        else:
            self.pos_widget.showwcs()
             
        # Update sliders
        self.blur_window.slider.setValue(int(self.image.r*10))
        self.frame_window.slider.setValue(self.frame)

        self.blur_window.slider.sliderReleased.connect(partial(self.image.blur,self.blur_window.slider.sliderPosition))
        self.frame_window.slider.valueChanged.connect(self.image.seek)

        self.frame_window.slider.setMaximum(self.image.n_frames-1)
        self.blur_window.slider.setMaximum(self.blur_max)

        # Update image label
        seen_text = ' (seen)' if self.image.seen else ''
        self.image_label.setText(f'{self.image.name} ({self.idx+1} of {self.N}){seen_text}')

        # Update menus
        self.update_mark_menu()

        if self.image.n_frames > 1:
            self.frame_action.setEnabled(True)
        else:
            self.frame_action.setEnabled(False)

        if self.image.wcs == None:
            self.settings_window.show_sexagesimal_box.setEnabled(False)
        else:
            self.settings_window.show_sexagesimal_box.setEnabled(True)

        # Set image as seen
        self.image.seen = True

    
    def update_comments(self):
        """Updates image comment with the contents of the comment box."""

        comment = self.comment_box.text()
        if not comment: comment = 'None'

        self.image.comment = comment

    def get_comment(self):
        """If the image has a comment, sets the text of the comment box to the image's comment."""

        if (self.image.comment == 'None'):
            self.comment_box.setText('')
        else:
            comment = self.image.comment
            self.comment_box.setText(comment)

    def update_categories(self):
        """Resets all category boxes to unchecked, then checks the boxes based on the current image's categories."""

        for box in self.category_boxes: box.setChecked(False)
        for i in self.image.categories:
            self.category_boxes[i-1].setChecked(True)

    def update_marks(self):
        """Redraws all marks in image."""

        # Update regular marks
        if self.image.duplicate == True:
            marks = self.image.dupe_marks
        else:
            marks = self.image.marks

        for mark in marks: 
            if mark not in self.image_scene.items():
                self.image_scene.mark(mark)

        # Update imageless marks
        if len(self.imageless_marks) > 0:

            # get imageless marks with coordinates in ra/dec
            marks_world = [mark for mark in self.imageless_marks if hasattr(mark,'_wcs_center')]
            
            # get imageless marks with coordinates in x/y
            marks_pix = [mark for mark in self.imageless_marks if hasattr(mark,'_center') ]

            # create list of ras/decs
            ra = [float(mark.wcs_center.ra) for mark in marks_world]
            dec = [float(mark.wcs_center.dec) for mark in marks_world]

            # create list of x/y
            x = [mark.center.x for mark in marks_pix]
            y = [mark.center.y for mark in marks_pix]

            # create pixcoords, converting ras/decs into x/y
            world_pix = WorldCoord(ra,dec).topix(self.image.wcs)
            pix_pix = PixCoord(x,y)

            # find which coordinates are inside the image
            world_filter = self.inview(world_pix.x,world_pix.y)
            pix_filter = self.inview(pix_pix.x,pix_pix.y)

            # add marks to image scene if it is inside the image
            for mark, viewable in zip(marks_world, world_filter):
                if viewable and (mark not in self.image_scene.items()):
                    mark.image = self.image
                    self.image_scene.mark(mark)
                    mark.image = None

            for mark, viewable in zip(marks_pix, pix_filter):
                if viewable and (mark not in self.image_scene.items()):
                    mark.image = self.image
                    self.image_scene.mark(mark)
                    mark.image = None
            
            self.update_mark_menu()

    def update_mark_menu(self):
        for path in io.markpaths():
            if path not in self.mark_menu.menus:
                self.mark_menu.menu_setup(path)
            else:
                self.mark_menu.update_menu(path)
                self.toggle_marks(path)
                self.toggle_mark_labels(path)

        menu_paths = self.mark_menu.menus.copy().keys()
        for path in menu_paths:
            if path not in io.markpaths():
                del self.mark_menu.menus[path]

    def update_colors(self,path):
        color = QColorDialog.getColor()
        if color.isValid():
            config.DEFAULT_COLORS[path] = color

            for item in self.image_scene.items():
                if hasattr(item,'dst'):
                    if (item.dst == path) and (item.g == 0):
                        pen = item.pen()
                        pen.setColor(color)
                        item.setPen(pen)
                        item.label.lineedit.setStyleSheet(
                            f"""background-color: rgba(0,0,0,0);
                                border: none; 
                                color: rgba{color.getRgb()}"""
                        )
                        
            self.mark_menu.update_color(path)
                
    def del_markfile(self, path):
        """Deletes a markfile."""

        if path == self.markfile.path:
            if self.image.duplicate == True:
                marks = [mark for mark in self.image.dupe_marks if mark.dst == path]
            else:
                marks = [mark for mark in self.image.marks if mark.dst == path]

            for mark in marks:
                self.image.undone_marks.append(mark)

                if mark in self.image_scene.items():
                    self.image_scene.rmmark(mark)
                
                if mark in self.image.marks:
                    self.image.marks.remove(mark)

                if mark in self.image.dupe_marks:
                    self.image.dupe_marks.remove(mark)
        
        else:
            for image in self.images:
                if image.duplicate == True:
                    marks = [mark for mark in image.dupe_marks if os.path.samefile(mark.dst, path)]
                    
                else:
                    marks = [mark for mark in image.marks if os.path.samefile(mark.dst, path)]

                for mark in marks:
                    if mark in self.image_scene.items():
                        self.image_scene.rmmark(mark)

                    if mark in image.marks:
                        image.marks.remove(mark)
                    
                    if mark in image.dupe_marks:
                        image.dupe_marks.remove(mark)

            os.remove(path)
                    
        imageless_marks = [mark for mark in self.imageless_marks if mark.dst == path]            

        for mark in imageless_marks:
            if mark in self.image_scene.items():
                self.image_scene.rmmark(mark)
            self.imageless_marks.remove(mark)

        self.update_mark_menu()
        
        self.save()
    
    def del_usermarks(self,mode='selected'):
        """Deletes marks, either the selected one or all."""
        
        if self.image.duplicate == True:
            marks = self.image.dupe_marks
        else:
            marks = self.image.marks

        if mode == 'all':
            selected_marks = [mark for mark in marks if mark.dst == self.markfile.path]
        elif mode == 'selected': 
            selected_marks = [mark for mark in marks 
                              if (mark.dst == self.markfile.path) and mark.isSelected()]
        elif mode == 'cursor':
            pix_pos = self.image_view.mouse_pix_pos(correction=False).toPointF()
            selected_marks = [mark for mark in marks if (mark is self.image_scene.itemAt(pix_pos, mark.transform()))
                              and (mark.dst == self.markfile.path)]
            
        for mark in selected_marks:
            self.image.undone_marks.append(mark)
            self.image_scene.rmmark(mark)
            marks.remove(mark)
        
        self.update_mark_menu()
            
        self.save()

    def toggle_randomize(self,state):
        """Updates the config file for randomization and reloads unseen images."""
        
        config.RANDOMIZE_ORDER = bool(state)
        config.update()

        names = [img.name for img in self.images]

        if not state: self.images = [self.images[i] for i in argsort(names)]

        else:
            unedited_names = [n for n in names if n not in self.order]

            rng = io.np.random.default_rng()
            rng.shuffle(unedited_names)

            randomized_names = self.order + unedited_names
            indices = [names.index(n) for n in randomized_names]
            self.images = [self.images[i] for i in indices]
     
        self.idx = self.images.index(self.image)

        self.update_images()
        self.update_marks()
        self.get_comment()
        self.update_categories()
        self.update_comments()

    def toggle_marks(self,path):
        """Toggles whether or not marks are shown."""

        if self.image.duplicate == True:
            marks = [mark for mark in self.image.dupe_marks if mark.dst == path]
        else:
            marks = [mark for mark in self.image.marks if mark.dst == path]

        marks += [mark for mark in self.imageless_marks if (mark.dst == path)]

        marks_enabled = self.mark_menu.marks_action(path).isChecked()
        labels_enabled = self.mark_menu.labels_action(path).isChecked()

        for mark in marks:
            if marks_enabled: 
                mark.show()
                self.mark_menu.labels_action(path).setEnabled(True)
                if labels_enabled: mark.label.show()
            else: 
                mark.hide()
                mark.label.hide()
                self.mark_menu.labels_action(path).setEnabled(False)

    def toggle_mark_labels(self,path):
        """Toggles whether or not mark labels are shown."""

        if self.image.duplicate == True:
            marks = [mark for mark in self.image.dupe_marks if mark.dst == path]
        else:
            marks = [mark for mark in self.image.marks if mark.dst == path]

        marks += [mark for mark in self.imageless_marks if mark.dst == path]

        marks_enabled = self.mark_menu.marks_action(path).isChecked()
        labels_enabled = self.mark_menu.labels_action(path).isChecked()

        for mark in marks:
            if marks_enabled and labels_enabled: mark.label.show()
            else: mark.label.hide()



"""
Copyright © 2025, UChicago Argonne, LLC

Full license found at _YOUR_INSTALLATION_DIRECTORY_/imgmarker/LICENSE
"""

"""This module contains custom PyQt widgets for Image Marker."""

from imgmarker.gui.pyqt import Qt, QLabel, QWidget, QHBoxLayout, QLineEdit, QFrame, QLineEdit, QSizePolicy, QFileDialog
import os

class QHLine(QFrame):
    def __init__(self):
        super(QHLine, self).__init__()
        self.setFrameShape(QFrame.Shape.HLine)
        self.setFrameShadow(QFrame.Shadow.Raised)
        self.setLineWidth(0)
        self.setMidLineWidth(1)
        self.setMinimumHeight(1)

class QVLine(QFrame):
    def __init__(self):
        super(QVLine, self).__init__()
        self.setFrameShape(QFrame.Shape.VLine)
        self.setFrameShadow(QFrame.Shadow.Raised)
        self.setLineWidth(0)
        self.setMidLineWidth(1)
        self.setMinimumHeight(1)

class PosWidget(QWidget):
    """
    Shows coordinates
    """
    def __init__(self):
        super().__init__()
        layout = QHBoxLayout()
        self.setLayout(layout)
        self.sep = QVLine()
       
        # pix text 
        self.x_text = QLineEdit()
        self._text_setup(self.x_text)
        self.y_text = QLineEdit()
        self._text_setup(self.y_text)
        
        self.x_label = QLabel()
        self._label_setup(self.x_label,'X:')
        self.y_label = QLabel()
        self._label_setup(self.y_label,'Y:')

        # wcs text 
        self.ra_text = QLineEdit()
        self._text_setup(self.ra_text)
        self.dec_text = QLineEdit()
        self._text_setup(self.dec_text)

        self.ra_label = QLabel()
        self._label_setup(self.ra_label,'RA:')
        self.dec_label = QLabel()
        self._label_setup(self.dec_label,'DEC:')

        # Add widgets to layout
        layout.addStretch(1)
        layout.addWidget(self.x_label)
        layout.addWidget(self.x_text)
        layout.addWidget(self.y_label)
        layout.addWidget(self.y_text)
        layout.addWidget(self.sep)
        layout.addWidget(self.ra_label)
        layout.addWidget(self.ra_text)
        layout.addWidget(self.dec_label)
        layout.addWidget(self.dec_text)
        layout.addStretch(1)

    def _text_setup(self,widget:QLineEdit):
        font = widget.font()
        font.setPointSizeF(font.pointSizeF()*0.75)
        widget.setFont(font)

        widget.setAlignment(Qt.AlignmentFlag.AlignCenter)
        widget.setReadOnly(True)
        widget.setFixedHeight(20)
        widget.setFixedWidth(90)
        widget.setSizePolicy(*[QSizePolicy.Policy.Fixed]*2)

    def _label_setup(self,label:QLabel,text:str):
        font = label.font()
        font.setPointSizeF(font.pointSizeF()*0.75)
        label.setFont(font)

        label.setText(text)
        label.setAlignment(Qt.AlignmentFlag.AlignCenter)
        label.setTextInteractionFlags(Qt.TextInteractionFlag.TextSelectableByMouse)
        label.setFixedHeight(20)

    def cleartext(self):
        self.x_text.setText('')
        self.y_text.setText('')
        self.ra_text.setText('')
        self.dec_text.setText('')

    def hidewcs(self):
        self.ra_label.hide()
        self.dec_label.hide()
        self.ra_text.hide()
        self.dec_text.hide()
        self.sep.hide()

    def showwcs(self):
        self.ra_label.show()
        self.dec_label.show()
        self.ra_text.show()
        self.dec_text.show()
        self.sep.show()
        
class RestrictedLineEdit(QLineEdit):
    def __init__(self,forbidden_keys:list=[]):
        super().__init__()
        self.forbidden_keys = forbidden_keys

    def keyPressEvent(self, a0):
        if not a0.key() in self.forbidden_keys: 
            return super().keyPressEvent(a0)
        
    def focusOutEvent(self, a0):
        if self.text() == '': self.setText(self.placeholderText())
        return super().focusOutEvent(a0)
    
class DefaultDialog(QFileDialog):
    def __init__(self,directory=os.path.expanduser('~')):
        #make this work with file dialog names on MacOS
        #default to user's home directory if a path isn't given. 
        # Create a QFileDialog instance
        super().__init__()
        self.setOption(QFileDialog.Option.DontUseNativeDialog, True)
        self.setFileMode(QFileDialog.FileMode.Directory)
        self.setDirectory(directory)
        self.closed = False

    def closeEvent(self, a0):
        self.closed = True
        return super().closeEvent(a0)
    
    def keyPressEvent(self, a0):
        if a0.key() == Qt.Key.Key_Escape: self.close()
        else: return super().keyPressEvent(a0)

    def selectedFiles(self):
        if self.closed: return None
        else: return super().selectedFiles()


"""
Copyright © 2025, UChicago Argonne, LLC

Full license found at _YOUR_INSTALLATION_DIRECTORY_/imgmarker/LICENSE
"""

from astropy.io.fits import open as _open
from astropy.io.fits import ImageHDU, PrimaryHDU, Header, HDUList

def open(path):
    _hdus = _open(path)

    if any(isinstance(hdu,ImageHDU) for hdu in _hdus):
        hdus = HDUList([_hdu for _hdu in _hdus if isinstance(_hdu,ImageHDU)])
    else:
        hdus = HDUList([_hdu for _hdu in _hdus if isinstance(_hdu,PrimaryHDU)])

    return hdus